Require Import vgtac Program List Arith Even Div2 Relations.
Require Import basis countable lang subs model.
Require Import Classical FunctionalExtensionality.
Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** ** automatically apply omega  *)
(*----------------------------------------------------------------------------*)

Hint Extern 3 (@eq nat _ _) => omega.
Hint Extern 3 (le _ _) => omega.
Hint Extern 3 (ge _ _) => omega.
Hint Extern 3 (lt _ _) => omega.
Hint Extern 3 (gt _ _) => omega.

(*----------------------------------------------------------------------------*)
(** ** Macros about Paco  *)
(*----------------------------------------------------------------------------*)

Tactic Notation "pdes" ident(H) := punfold H; destruct H; pclearbot.
Tactic Notation "pinv" ident(H) := punfold H; inv H; pclearbot.
Tactic Notation "pdepdes" ident(H) := punfold H; dependent destruction H; pclearbot.

Lemma or_false: forall P, P \/ False -> P.
Proof. intuition. Qed.

(*----------------------------------------------------------------------------*)
(** ** Injectivities *)
(*----------------------------------------------------------------------------*)

Lemma unflex_inj : forall t1 t2, unflex t1 = unflex t2 -> t1 = t2.
Proof.
  by destruct t1; intros; destruct t2; inversion H; try rewrite <- H1, <- H2.
Qed.

Lemma exp_of_val_inj: forall v v', exp_of_val v = exp_of_val v' -> v = v'.
Proof. 
  induction v; destruct v'; intross; clarify; try by rewrite (IHv _ H).
  by rewrite (IHv1 _ H1), (IHv2 _ H).
Qed.

Lemma e_of_inj :
  forall v v', e_of v = e_of v' -> v_of v = v_of v'.
Proof. by intros; destruct v, v'; subst; apply exp_of_val_inj in H; simpls; subst. Qed.

Lemma v_of_inj :
  forall v v', v_of v = v_of v' -> v = v'.
Proof. by intros; destruct v, v'; simpls; subst. Qed.

Ltac simpl_inj :=
  repeat match goal with 
  | H : unflex _ = unflex _ |- _ => 
      try apply unflex_inj in H; let P := type of H in fold (__HID__ P) in H
  | H : v_of _ = v_of _ |- _ => 
      try apply v_of_inj in H; let P := type of H in fold (__HID__ P) in H
  | H : _ = _ |- _ => 
      try apply e_of_inj in H; let P := type of H in fold (__HID__ P) in H
  end; unfold __HID__ in *.
Ltac simpl_injs := simpl_inj; simpl v_of in *.

(*----------------------------------------------------------------------------*)
(** ** Miscellaneous 'library' definitions and lemmas *)
(*----------------------------------------------------------------------------*)

Definition swap {A B} (x : A * B) := (snd x, fst x).

Lemma list_assoc_in: forall A B eqdec l a b
    (FOUND: @list_assoc A B eqdec a l = Some b),
  In (a,b) l.
Proof.
induction l; intros; simpl; [discriminate|].
destruct a; simpl in *; destruct (eqdec a a0); subst; eauto.
inversion FOUND; subst; eauto.
Qed.

Lemma list_minus_emp_spec :
  forall A myeq l1 l2,
    (list_minus (A:=A) myeq l1 l2 = nil) <-> (forall x, In x l1 -> In x l2).
Proof.
  intros; induction l1 as [|a l1 IH]; simpl; intros; try tauto.
  generalize (list_mem_spec myeq a l2); destruct list_mem;
  intuition; congruence. 
Qed.

Lemma subst_closed_gg : 
  forall (f : (nat*val)*(nat*val) -> exp) A (gg : list (termvar*A*((nat*val)*(nat*val)))) 
    (H: forall x, In x gg -> fv_exp (f (snd x)) = []),
  subst_closed (map (fun x=>(fst(fst x), f(snd x))) gg).
Proof. induction gg; intross; auto. Qed.

(*----------------------------------------------------------------------------*)
(** ** Helper lemmas for proving equalities up to proof irrelevance *)
(*----------------------------------------------------------------------------*)

Lemma sig_eq : 
  forall A (P: A -> Prop) (x x' : sig P), proj1_sig x = proj1_sig x' -> x = x'.
Proof. destruct x, x'; simpl; intro; subst; f_equal; apply proof_irr. Qed.

Lemma heap_eq : forall h1 h2, (forall x, heap_f h1 x = heap_f h2 x) -> h1 = h2.
Proof. 
  intros [] [] ?; simpl in *; apply functional_extensionality in H.
  subst; f_equal; apply proof_irr.
Qed. 

Lemma sref_eq :
  forall (s s': wst wref) (EQ: forall ty l l', `s ty l l' <-> `s' ty l l'), s = s'. 
Proof.
  by intros; eapply (@sig_eq _ _ s); eauto using @functional_extensionality, prop_ext.
Qed.

(*----------------------------------------------------------------------------*)
(** ** Properties of heaps *)
(*----------------------------------------------------------------------------*)

Definition hunion' (h1 h2 : heap) l :=
  match h1 l with Some v1 => Some v1 | _ => h2 l end.

Definition hunion (h1 h2 : heap) : heap.
  refine (@Build_heap (hunion' h1 h2) _).
  intros.
  destruct (heap_findom h1) as (dom1, ?); destruct (heap_findom h2) as (dom2, ?).
  exists (app dom1 dom2); intros l ?.
  by unfold hunion'; case_eq (h1 l); intros; apply in_or_app; [left | right]; eauto.
Defined.

Definition hrestrict' (h : heap) (dom : location -> Prop) l :=
  match LEM (dom l) with
  | left _ => h l
  | right _ => None
  end.

Definition hrestrict (h  : heap )(dom : location -> Prop) : heap.
  refine (@Build_heap (hrestrict' h dom) _).
  destruct (heap_findom h) as (hdom, ?); exists hdom; unfold hrestrict'.
  by intro l; case_eq (LEM (dom l)); intros; eauto.
Defined.

Lemma hplus_det : 
  forall h1 h2 h (H: hplus h1 h2 h) h' (H': hplus h1 h2 h'), h' = h.
Proof. 
  intros; apply heap_eq; intro x.
  destruct (H x); destruct (H' x); des; congruence.
Qed.

Lemma hplus_hemp_l : forall h, hplus hemp h h.
Proof. vauto. Qed.

Lemma hplus_hemp_r : forall h, hplus h hemp h.
Proof. vauto. Qed.

Lemma hplus_hemp_lE : forall h h', hplus hemp h h' -> h = h'.
Proof.
  intros; apply heap_eq; intro x; destruct (H x); des; clarify; simpls; congruence.
Qed.

Lemma hplus_hemp_rE : forall h h', hplus h hemp h' -> h = h'.
Proof.
  intros; apply heap_eq; intro x; destruct (H x); des; clarify; simpls; congruence.
Qed.

Lemma hplus_hemp_inv : forall h h', hplus h h' hemp -> h = hemp /\ h' = hemp.
Proof.
  by intros; split; eapply heap_eq; intros x; specialize (H x); intuition.
Qed.

Lemma hplus_comm : forall h1 h2 h, hplus h1 h2 h -> hplus h2 h1 h.
Proof. by red; intros; edestruct H; vauto. Qed.

Lemma hplus_assoc_l : 
  forall h1 h2 h12 (R: hplus h1 h2 h12) h3 h (R': hplus h12 h3 h),
    exists h23, hplus h1 h23 h /\ hplus h2 h3 h23. 
Proof.
  intros; eexists (hunion h2 h3); split; intro x; simpl; unfold hunion';
    specialize (R x); specialize (R' x); destruct (h2 x); des; vauto;
    constructor (split; congruence). 
Qed.

Lemma hplus_assoc_r : 
  forall h2 h3 h23 (R: hplus h2 h3 h23) h1 h (R': hplus h1 h23 h),
    exists h12, hplus h1 h2 h12 /\ hplus h12 h3 h. 
Proof.
  intros; eexists (hunion h1 h2); split; intro x; simpl; unfold hunion';
    specialize (R x); specialize (R' x); destruct (h1 x); des; vauto;
    constructor (split; congruence). 
Qed.

Lemma hplus_none :
  forall h1 h2 h (P: hplus h1 h2 h) loc (N: h loc = None),
    h1 loc = None /\ h2 loc = None.
Proof.
  intros; specialize (P loc); simpls; des; split; congruence.
Qed.

Lemma hplus_some1 :
  forall h1 h2 h (P: hplus h1 h2 h) loc v (N: h1 loc = Some v),
    h loc = Some v.
Proof. intros; specialize (P loc); simpls; des; congruence. Qed.

Lemma hplus_some2 :
  forall h1 h2 h (P: hplus h1 h2 h) loc v (N: h2 loc = Some v),
    h loc = Some v.
Proof. intros; specialize (P loc); simpls; des; congruence. Qed.

Lemma hplus_inl : forall h1 h2 h12 (HP : hplus h1 h2 h12) l v (EH : h1 l = Some v),
  h12 l = Some v.
Proof.
  by intros; specialize (HP l); des; [rewrite HP0 in EH | rewrite HP].
Qed.

Lemma hplus_inr : forall h1 h2 h12 (HP : hplus h1 h2 h12) l v (EH : h2 l = Some v),
  h12 l = Some v.
Proof.
  by intros; specialize (HP l); des; [rewrite HP | rewrite HP0 in EH].
Qed.

Lemma hplus_assoc : forall h1 h2 h3 h12 h23 h123
    (HP12 : hplus h1 h2 h12)
    (HP123 : hplus h12 h3 h123)
    (HP23 : hplus h2 h3 h23),
  hplus h1 h23 h123.
Proof.
  intros; assert (HP := hplus_assoc_l HP12 HP123); des.
  by rewrite (hplus_det HP0 HP23).
Qed.

Lemma hplus_disjoint : forall h1 h2 h12 (HP : hplus h1 h2 h12) l
    v1 (IN1 : h1 l = Some v1) v2 (IN2 : h2 l = Some v2),
  False.
Proof.
  by intros; destruct (HP l); des; [rewrite H0 in IN1 | rewrite H0 in IN2].
Qed.

Lemma hunion_inl : forall (h1 h2 : heap) l v (INL : h1 l = Some v),
  hunion h1 h2 l = Some v.
Proof.
  by intros; simpl; unfold hunion'; rewrite INL.
Qed.

Lemma hunion_inr : forall (h1 h2 : heap) l v (INR : h2 l = Some v) h12 (DJ : hplus h1 h2 h12),
  hunion h1 h2 l = Some v.
Proof.
  intros; simpl; unfold hunion'; specialize (DJ l); revert DJ.
  by rewrite INR; case_eq (h1 l); intros; des; vauto.
Qed.

Hint Immediate hplus_hemp_l hplus_hemp_r hplus_comm. 

Ltac clarify_heap :=
  clarify; repeat match goal with
  | H: hplus hemp _ _ |- _ => apply hplus_hemp_lE in H; clarify
  | H: hplus _ hemp _ |- _ => apply hplus_hemp_rE in H; clarify
  | H: hplus ?x ?y _, H' : hplus ?x ?y _ |- _ => apply (hplus_det H) in H'; clarify
  end.

(*----------------------------------------------------------------------------*)
(** ** Properties of the underlying untyped language *)
(*----------------------------------------------------------------------------*)

Inductive reduces h e : heap -> exp -> Prop :=
| reduces_nil : reduces h e h e
| reduces_step : forall h1 e1 h2 e2
    (STEP: reduce h e h1 e1)
    (RED: reduces h1 e1 h2 e2),
  reduces h e h2 e2.
Hint Resolve reduces_step.

Inductive diverges_gen diverges h e : Prop :=
| diverges_step : forall h' e' 
    (STEP: reduce h e h' e')
    (DIV: diverges h' e': Prop), 
  diverges_gen diverges h e.
Hint Constructors diverges_gen.

Definition diverges h e := paco2 diverges_gen bot2 h e.
Hint Unfold diverges.
Lemma diverges_gen_mon: monotone2 diverges_gen. Proof. pmonauto. Qed.
Hint Resolve diverges_gen_mon : paco.

Inductive gets_stuck h e : Prop :=
_gets_stuck h' e'
  (RED : reduces h e h' e')
  (STUCK : forall h'' e'', ~ reduce h' e' h'' e'')
  (NOTVAL: ~ is_value_of_exp e').
Hint Constructors gets_stuck.

Definition is_val e := exists ve : voe, e = e_of ve.
Hint Unfold is_val.

Lemma voe_is_val: forall e, is_value_of_exp e -> is_val e.
Proof.
  i; induction e; eauto; ss; des
  ; try edestruct IHe; try edestruct IHe1; try edestruct IHe2; subst; eauto.
Qed.  

Lemma val_is_voe: forall e, is_val e -> is_value_of_exp e.
Proof.
  i; rdes H; subst; destruct ve; subst; s; induction v_of; s; eauto.
Qed.

Lemma app_cont_assoc: forall k1 k2 e,
   app_cont (k1 ++ k2) e = app_cont k2 (app_cont k1 e).
Proof. by intros; unfold app_cont; rewrite fold_left_app. Qed.

Lemma int_app_stuck: forall h1 h2 i e, ~ reduce h1 (Eapp (Vint i) Vunit) h2 e.
Proof.
  intros; inv FH; destruct SK; inv H1; simpls. 
  - by inv RED; destruct SK; inv H2.
  - by simpl_injs; subst; inv RED; destruct SK; inv H2.
Qed.

(** Values cannot reduce. *)

Lemma not_reduce_val :
   forall h ve h' e', ~ reduce h (e_of ve) h' e'.
Proof.
  intros; remember (e_of ve) as e; revert ve Heqe.
  induction FH; try by intros []; intros; subst; destruct v_of.
  intros.
  cut (exists v: val, e = v); [by intro; des; eauto|].
  revert ve Heqe; clear.
  destruct SK; simpls; intros; eauto; 
   try by destruct ve; subst; destruct v_of; simpls; clarify; eauto.
Qed.

Lemma reduce_not_val :
  forall h e h' e', reduce h e h' e' -> ~ is_value_of_exp e.
Proof. i; apply voe_is_val in FH; rdes FH; eapply not_reduce_val; subst; eauto. Qed.

(** The operational semantics is deterministic. *)

Lemma reduce_det :
   forall h e h1 e1 (R1: reduce h e h1 e1) h2 e2 (R2: reduce h e h2 e2),
     h1 = h2 /\ e1 = e2.
Proof.
  induction 1; inversion 1; clarify;
  try (destruct SK; simpls; clarify);
  try (destruct SK0; simpls; clarify; try destruct (IHR1 _ _ RED); clarify);
  try (by destruct (reduce_not_val RED); eauto using val_is_voe);
  try (by destruct (reduce_not_val R1); eauto using val_is_voe);
  simpl_inj; simpl_inj; clarify;
  try (by exfalso; eauto).
  split; try done.
  apply heap_eq; intro x; generalize (f_equal (fun f => f x) H0); simpl.
  by destruct Peano_dec.eq_nat_dec.
Qed.

Lemma reduce_app_cont_det:
  forall h K e h' e' h'' e''
    (RC: reduce h (app_cont K e) h' e')
    (RED: reduce h e h'' e''),
   e' = app_cont K e''.
Proof. intros; apply (reduce_det RC (Rctx RED K)). Qed.

Lemma reduces_val_det :
   forall h e h1 e1 (R1: reduces h e h1 e1) (v1: val) (E1 : e1 = v1)
              h2 e2 (R2: reduces h e h2 e2) (v2: val) (E2 : e2 = v2),
     h1 = h2 /\ v1 = v2.
Proof.
  intros until e1; induction 1; intros; subst.
    depdes R2; subst; eauto using exp_of_val_inj.
    by exfalso; eapply not_reduce_val; eauto.
  depdes R2; subst.
    by exfalso; eapply not_reduce_val; eauto.
  destruct (reduce_det STEP STEP0); subst.
  exploit IHR1; try apply R2; eauto; i; des; eauto.
Qed.

Lemma reduces_stuck_det :
   forall h e h1 e1 (R1: reduces h e h1 e1) (NRS1 : forall h' e', ~ reduce h1 e1 h' e')
                    (NV1 : ~ is_value_of_exp e1)
              h2 e2 (R2: reduces h e h2 e2) (NRS2 : forall h' e', ~ reduce h2 e2 h' e')
                    (NV2 : ~ is_value_of_exp e2),
     h1 = h2 /\ e1 = e2.
Proof.
  intros until e1; induction 1; intros.
    by destruct R2; subst; auto; exfalso; eapply NRS1; eauto.
  destruct R2; subst. 
    by exfalso; eapply NRS2; eauto.
  destruct (reduce_det STEP STEP0); subst.
  exploit IHR1; try apply R2; eauto; i; des; eauto.
Qed.

Lemma value_cont : forall k e (V: is_value_of_exp (app_cont k e)), is_value_of_exp e.
Proof.
  i; apply voe_is_val in V; apply val_is_voe; revert e V.
  induction k; eauto; i; ss; eapply IHk in V.
  rdes V; destruct ve; subst; destruct a, v_of; ss; inv V; eauto.
Qed.

Lemma diverges_reduces : 
  forall h e h' e' (R: reduces h e h' e') (D: diverges h e),
    diverges h' e'.
Proof. 
  induction 1; subst; intross.
  pdes D; eapply (reduce_det STEP) in STEP0; des; clarify; eauto.
Qed.

Lemma diverges_no_stuck: 
  forall h e, diverges h e -> ~ gets_stuck h e.
Proof.
  intros; destruct FH.
  induction RED; pinv H; try destruct (reduce_det STEP STEP0); subst; eauto.
Qed.

Lemma reduces_val_no_stuck: 
  forall h e h' (v: val), reduces h e h' v -> ~ gets_stuck h e.
Proof.
  intros; destruct FH. 
  move RED at top; revert_until RED; induction RED; i; subst
  ; inv H; try destruct (reduce_det STEP STEP0); subst
  ; eauto using not_reduce_val, val_is_voe.
Qed.

Lemma reduces_val_no_diverges :
 forall h e h' (v: val), reduces h e h' v -> ~ diverges h e.
Proof.
  intros; eapply diverges_reduces in H; eauto.
  by pinv H; eauto using not_reduce_val.
Qed.

Lemma reduces_trans : 
  forall h e h' e' (R: reduces h e h' e') 
    h'' e'' (R': reduces h' e' h'' e''),
    reduces h e h'' e''.
Proof. induction 1; subst; [|i; eapply reduces_step; s]; eauto. Qed.

Lemma app_scont_eq: forall sk sk' e e'
    (EQ: appctx_scont_exp sk e = appctx_scont_exp sk' e')
    (NOV: ~ is_value_of_exp e)
    (NOV': ~ is_value_of_exp e'),
  sk = sk' /\ e = e'.
Proof.
  i; assert (NOV1 := fun x => NOV (val_is_voe x)); clear NOV.
  assert (NOV2 := fun x => NOV' (val_is_voe x)); clear NOV'.
  by intros; destruct sk, sk'; try inv EQ;
   try (exfalso; first [by apply NOV1; eauto|by apply NOV2; eauto]); 
   simpl_injs; subst.
Qed.

Definition is_atom e := 
  forall sk e' (EQ: e = appctx_scont_exp sk e') (NV: ~ is_value_of_exp e'), False.

Lemma atom_generalize: forall e (ATM: is_atom e) 
    k e' (EQ: e = app_cont k e') (NV: ~ is_value_of_exp e'),
  k = [].
Proof.
  induction k; eauto; intros.
  exploit (IHk _ EQ); eauto using (value_cont [a]).
  intros; subst; exfalso; eapply ATM; simpls. 
Qed.

Lemma rev_revert: forall A l (P: list A -> Prop) (ASM: forall l, P (rev l)), P l.
Proof. by intros; specialize (ASM (rev l)); rewrite rev_involutive in ASM. Qed.

Lemma app_cont_r_eq : forall k k' e e'
    (EQ: app_cont_r k e = app_cont_r k' e')
    (NOV: ~ is_value_of_exp e)
    (NOV': ~ is_value_of_exp e')
    (ATM1 : is_atom e)
    (ATM2 : is_atom e'),
  k = k' /\ e = e'.
Proof.
  intros; assert (ATOM1 := atom_generalize ATM1); assert (ATOM2 := atom_generalize ATM2).
  clear ATM1 ATM2; revert_until k; induction k; intros.
  - destruct k'; eauto.
    exploit (ATOM1 (rev (s::k')) e'); eauto. 
    : by rewrite <-app_cont_r_reverse; eauto.
    intro X; assert (Y:= rev_involutive (s::k')); rewrite X in Y; inv X.
  - destruct k'.
    + exploit (ATOM2 (rev (a::k)) e); eauto. 
      : by rewrite <-app_cont_r_reverse; eauto.
      intro X; assert (Y:= rev_involutive (a::k)); rewrite X in Y; inv X.
    + simpl in EQ; eapply app_scont_eq in EQ.
      * des; subst; eapply IHk in EQ0; des; subst; eauto. 
      * rewrite app_cont_r_reverse; eauto using value_cont.
      * rewrite app_cont_r_reverse; eauto using value_cont.
Qed.

Lemma app_cont_eq : forall k k' e e'
    (EQ: app_cont k e = app_cont k' e')
    (NOV: ~ is_value_of_exp e)
    (NOV': ~ is_value_of_exp e')
    (ATM1 : is_atom e)
    (ATM2 : is_atom e'),
  k = k' /\ e = e'.
Proof.
  intro k; apply (rev_revert k); clear k; intro k.
  intro k'; apply (rev_revert k'); clear k'; intro k'.
  setoid_rewrite <- app_cont_r_reverse.
  by intros; exploit app_cont_r_eq; eauto; intro; des; subst.
Qed.

Lemma reduce_app_cont_inv: forall K h e h' e' 
    (R : reduce h (app_cont K e) h' e') 
    (EXP: ~ is_value_of_exp e),
  exists e'', reduce h e h' e''.
Proof.
  induction K; eauto.
  intros; simpls; exploit IHK; eauto using (value_cont [a]).
  intros; des; depdes x0 
  ; try by destruct a; depdes x; ss; exfalso; apply EXP; esplits; eauto using val_is_voe.
  apply app_scont_eq in x; des; subst; eauto using reduce_not_val.
Qed.

Lemma not_reduce_cont : forall k h e h' e'
    (NR : forall h' e', ~ reduce h e h' e') 
    (NV : ~ is_value_of_exp e),
  ~ reduce h (app_cont k e) h' e'.
Proof. intros; edestruct reduce_app_cont_inv; eauto. Qed.

Lemma lreds_reds: forall n h e n1 h1 e1 
    (REDS: lreduces n h e n1 h1 e1),
  reduces h e h1 e1.
Proof.
  i; induction REDS; [econs|depdes STEP]; eauto.
Qed.

Lemma reds_lreds: forall n h e h1 e1 
    (REDS: reduces h e h1 e1),
  lreduces n h e n h1 e1.
Proof.
  i; induction REDS; [econs|depdes STEP]; eauto.
Qed.

Lemma div_ldiv: forall n h e (DIV: diverges h e), ldiverges n h e.
Proof. 
  pcofix CIH; i; pdes DIV; pfold; eauto.
Grab Existential Variables. exact 0.
Qed.

Lemma ldiv_div: forall n h e (DIV: ldiverges n h e), diverges h e.
Proof.
  pcofix CIH. 
  intro n; generalize (le_n n); generalize n at -2; induction n.
  - i; pdes DIV; depdes STEP; [omega|pfold; eauto].
  - i; pdes DIV; depdes STEP.
    + eapply IHn, DIV; omega.
    + pfold; eauto.
Qed.

Lemma lreds_trans : 
  forall n h e n' h' e' (R: lreduces n h e n' h' e') 
    n'' h'' e'' (R': lreduces n' h' e' n'' h'' e''),
    lreduces n h e n'' h'' e''.
Proof. induction 1; subst; [|i; eapply lreduces_step; s]; eauto. Qed.

Lemma ldiv_mon: forall n n' h e (DIV: ldiverges n h e) (LE: n' >= n), ldiverges n' h e.
Proof.
  i; pdes DIV; pfold; depdes STEP; eauto. 
  econs; eauto; econs; omega.
Qed.

Lemma lreds_mon: forall n h e n1 h1 e1
    (DIV: lreduces n h e n1 h1 e1) 
    n' (GE: n' >= n) n1' (LE: n1' <= n1), 
  lreduces n' h e n1' h1 e1.
Proof.
  induction 1; i.
  - assert (n1' = n' \/ n1' < n') by omega.
    des; subst; eauto using lreduces_nil.  
  - econs; [|by eapply IHDIV; eauto].
    depdes STEP; [econs; omega|eauto].
Qed.

Lemma lreds_ldiv : 
  forall r n h e n' h' e' (R: lreduces n h e n' h' e') (D: paco3 ldiverges_gen r n' h' e'),
    paco3 ldiverges_gen r n h e.
Proof. induction 1; subst; i; eauto. Qed.

Lemma lred_cont : 
  forall n h e n' h' e' (D: lreduce n h e n' h' e') K,
    lreduce n h (app_cont K e) n' h' (app_cont K e').
Proof. i; depdes D; eauto. Qed.

Lemma lreds_cont :
  forall n h e n' h' e' (D: lreduces n h e n' h' e') K,
    lreduces n h (app_cont K e) n' h' (app_cont K e').
Proof. 
  induction 1; subst; eauto using lreduces_nil, lred_cont.
Qed.

Lemma diverges_cont : 
  forall h e (D: diverges h e) K, diverges h (app_cont K e).
Proof. 
  pcofix X; i; pdes D; pfold; econs; eauto. 
Qed.

Lemma lreds_plus: forall n h e n1 h1 e1,
  lreduces n h e n1 h1 e1 -> forall n', lreduces (n+n') h e (n1+n') h1 e1.
Proof.
  induction 1; eauto using lreduces_nil.
  i; econs; eauto; depdes STEP; eauto.
Qed.

Lemma reds_lreds_val: forall n h e n1 h1 e1 
    (E: ~is_value_of_exp e) (V: is_value_of_exp e1),
  reduces h e h1 e1 -> lreduces n h e n1 h1 e1.
Proof.
  i; move H at top; revert_until H; induction H; i.
  - exfalso; eauto.
  - destruct (classic(is_value_of_exp e1)).
    + apply voe_is_val in V; apply voe_is_val in H0; rdes V; rdes H0.
      destruct ve, ve0; ss; subst.
      eapply reduces_val_det in H; try apply reduces_nil; eauto.
      des; subst; eauto using lreduces_nil.
    + eauto.
Grab Existential Variables. eauto.
Qed.

Lemma lred_lreds: forall n h e n' h' e',
  lreduce n h e n' h' e' -> lreduces (n - 1) h e n' h' e'.
Proof.
  i; destruct H.
  - eapply lreds_mon; eauto using lreduces_nil; omega.
  - eauto using lreduces_nil.
Qed.

(*----------------------------------------------------------------------------*)
(** ** Tactic and Lemmas for language syntax (closedness, substitution, etc)  *)
(*----------------------------------------------------------------------------*)

Tactic Notation "simpl_clo" int_or_var(n) := 
  repeat (repeat(
    match goal with 
    | [ |- _ /\ _ ] => split
    | [ |- _ ++ _ = [] ] => eapply app_nil_invl
    | [H : _ ++ _ = [] |- _] => destruct (@app_nil_invr _ _ _ H); try clear H
    | [ |- fv_cont(_ ++ _) = [] ] => eapply fv_cont_appl
    | [H :fv_cont(_ ++ _) = [] |- _] => eapply fv_cont_appr in H
    | [ |- list_minus eq_termvar (_ ++ _) _ = [] ] => rewrite list_minus_app
    | [H : list_minus eq_termvar (_ ++ _) _ = [] |- _] => rewrite list_minus_app in H
    end); eauto n; simpl_inj; simpls; subst; des).

Lemma subst_exp_var: forall sub x (e: exp)
    (IN: list_assoc eq_termvar x sub = Some e),
  subst_exp sub (Evar x) = e.
Proof.
  induction sub; ss; i.
  destruct a; destruct eq_termvar; eauto.
  subst t; inv IN.
Qed.

Lemma subst_exp_val: forall sub (e: exp)
    (VAL: is_value_of_exp e),
  is_value_of_exp (subst_exp (sv2e sub) e).
Proof.
  induction e; s; i; des; eauto; clear VAL.
  induction sub; s; eauto.
  destruct eq_termvar; eauto.
  destruct a; subst; eauto using val_is_voe.
Qed.

Lemma subst_exp_val_inv: forall (e: exp) sub
    (VAL: is_value_of_exp (subst_exp (sv2e sub) e)),
  is_value_of_exp e.
Proof.
  induction e; s; i; des; eauto.
Qed.

(*----------------------------------------------------------------------------*)
(** ** Properties of the typing judgments *)
(*----------------------------------------------------------------------------*)

Lemma hastype_wfenv :
  forall D G e ty (HT: hastype D G e ty), wfenv D G.
Proof. induction 1; eauto; inv H. Qed.

Lemma hastype_wftyp :
  forall D G e ty (HT: hastype D G e ty), wftyp D G ty.
Proof.
  induction 1; eauto; vauto;
  repeat (rewrite ?list_minus_app in *; des; match goal with
    | H : _ ++ _ = nil |- _ => eapply app_eq_nil in H
    | H : _ |- _ => rewrite H; []; simpl
    | H : list_assoc ?myeq ?x ?l = _ |- _ =>
      pose proof (list_assoc_spec myeq x l); destruct (list_assoc myeq x l); clarify
    | H : wftyp _ _ _ |- _ => inversion_clear H; simpl in *; clarify; eauto
    | |- wftyp _ _ _ => constructor; simpl; eauto using lc_tsubst1, lc_tsubst1D
    | H : wfenv _ _ |- _ => inversion_clear H; simpl in *; clarify; eauto
    | |- list_minus _ _ _ = [] => apply list_minus_emp_spec; intros; eauto
    | H : list_minus _ _ _ = [] |- In _ _ =>
      solve [eapply list_minus_emp_spec; [eapply H|eauto using in_ftn_tsubst1_mon] ]
    | H : In _ (ftn_typ (tsubst1 _ _ _)) |- _ => apply in_ftn_tsubst1 in H; simpl in H; destruct H; eauto
  end).
Qed.

Lemma wfctx_wfenv :
  forall C D G ty D' G' ty' (HT: wfctx C D G ty D' G' ty'), wfenv D G.
Proof. induction 1; eauto; inv H. Qed.

Lemma wfctx_wftyp :
  forall C D G ty D' G' ty' (HT: wfctx C D G ty D' G' ty'), wftyp D' G' ty'.
Proof.
  induction 1; eauto; vauto;
  repeat (rewrite ?list_minus_app in *; des; match goal with
    | H : _ ++ _ = nil |- _ => eapply app_eq_nil in H
    | H : _ |- _ => rewrite H; []; simpl
    | H : list_assoc ?myeq ?x ?l = _ |- _ =>
      pose proof (list_assoc_spec myeq x l); destruct (list_assoc myeq x l); clarify
    | H : wftyp _ _ _ |- _ => inversion_clear H; simpl in *; clarify; eauto
    | |- wftyp _ _ _ => constructor; simpl; eauto using lc_tsubst1, lc_tsubst1D
    | H : wfenv _ _ |- _ => inversion_clear H; simpl in *; clarify; eauto
    | |- list_minus _ _ _ = [] => apply list_minus_emp_spec; intros; eauto
    | H : list_minus _ _ _ = [] |- In _ _ =>
      solve [eapply list_minus_emp_spec; [eapply H|eauto using in_ftn_tsubst1_mon] ]
    | H : In _ (ftn_typ (tsubst1 _ _ _)) |- _ => apply in_ftn_tsubst1 in H; simpl in H; destruct H; eauto
    | H : phastype _ _ _ _ |- _ => apply hastype_erase, hastype_wftyp in H; simpl in H; eauto
  end).
Qed.

(*----------------------------------------------------------------------------*)
(** * wvalmin *)
(*----------------------------------------------------------------------------*)

Inductive wvalmin (w : world) (s: wst w) (ty: flextyp) n1 (v1: val) n2 (v2: val) : Prop := 
| wvalmin_intro r (REL: wval w () s r ty n1 v1 n2 v2) (MON: r <5= wvalmin w s).
Hint Constructors wvalmin.

Lemma wvalmin_glob: forall (w : world), wglob _ (wvalmin w).
Proof.
  split; [|split]; repeat intro.
  + induction V; eapply wval_closed; eauto.
  + revert s' n1 n1' EXT LEN LEN'; induction R; eauto using wval_mon.
  + split; [|split]; unnw; intros; try destruct sp; try destruct R; eauto using wval_mon.
Qed.

Lemma wvalmin_init: forall w r (G: wglob _ r), wvalmin w <6= r.
Proof.
  intros; rdes G; rhdes GSUP; induction PR; eauto using wval_mon.
Qed.

(*----------------------------------------------------------------------------*)
(** * Language-independent lemmas about worlds *)
(*----------------------------------------------------------------------------*)

Program Definition wid : world_local := 
  wsingle (fun _ _ _ _ _ _ => False) (fun _ h h' => h = hemp /\ h' = hemp) _ _ _ _.

Lemma wstable_id : wstable wid.
Proof. repeat red; eauto. Qed.

Lemma winhabited_id : winhabited (wlift wid).
Proof. exists (sref0,()); repeat exists hemp; eby repeat split; eauto. Qed.

Lemma wconsistent_id : wconsistent (wlift wid).
Proof. red; inversion 2; intros; simpls; des; clarify. Qed.

Hint Resolve  wstable_id winhabited_id wconsistent_id.

Lemma wstuck_mon: forall rf rv rf' rv' ty n1 e1 n2 e2 
    (WS: wstuck rf rv ty n1 e1 n2 e2) (LEF: rf <5= rf') (LEV: rv <5= rv'),
  wstuck rf' rv' ty n1 e1 n2 e2.
Proof. by intros; inv WS; [econstructor 1|econstructor 2]; eauto using vclos_mon. Qed.

Lemma sref0_hemp: forall s r (HR: wheap wref () s r hemp hemp), s = sref0.
Proof.
  intros; apply sref_eq; intuition; repeat red in HR; des.
  by specialize (HSAT _ _ _ H); des.
Qed.

Lemma wrefst_inj1 : forall (s : wst wref) l1
    ty l2 (IN : `s ty l1 l2)
    ty' l2' (IN' : `s ty' l1 l2'),
  ty = ty' /\ l2 = l2'.
Proof.
  intros; destruct s; simpls.
  by rrdes w; edestruct w0 as (A, ?); [apply IN | apply IN' | specialize (A eq_refl)].
Qed.

Lemma wrefst_inj2 : forall (s : wst wref) l2
    ty l1 (IN : `s ty l1 l2)
    ty' l1' (IN' : `s ty' l1' l2),
  ty = ty' /\ l1 = l1'.
Proof.
  intros; destruct s; simpls.
  by rrdes w; edestruct w0 as (?, A); [apply IN | apply IN' | specialize (A eq_refl)].
Qed.

Implicit Arguments wrefst_inj1 [s l1 ty l2 ty' l2'].
Implicit Arguments wrefst_inj2 [s l2 ty l1 ty' l1'].

Definition wdisj (w1 w2 : world_local) := ndisj (wnam w1) (wnam w2).

Lemma wdisj_comm : 
  forall w1 w2, wdisj w1 w2 -> wdisj w2 w1.
Proof. eby red; red; intros; eapply H. Qed.

Lemma wdisj_prod_l : 
  forall w1 w2 w (D: wdisj w1 w) (D': wdisj w2 w), 
  wdisj (wprod w1 w2) w.
Proof. by intros; intros n []; [apply D | apply D']. Qed.

Lemma wdisj_prod_r : 
  forall w1 w2 w (D: wdisj w w1) (D': wdisj w w2),
  wdisj w (wprod w1 w2).
Proof. intros; auto using wdisj_prod_l, wdisj_comm. Qed.

Hint Resolve wdisj_prod_l wdisj_prod_r.
Hint Immediate wdisj_comm.

(*----------------------------------------------------------------------------*)
(** ** Basic properties of equivalence definitions *)
(*----------------------------------------------------------------------------*)

Lemma wstuck_prod:
  forall (w1 : world_local) w2 sp s r r' ty n1 e1 n2 e2
    (EREL : wstuck (wval (wprod w1 w2) sp s r) r' ty n1 e1 n2 e2),
  wstuck (wval w1 sp (fst s) r) r' ty n1 e1 n2 e2 \/ 
  wstuck (wval w2 sp (snd s) r) r' ty n1 e1 n2 e2.
Proof. intros; inv EREL; inv FREL; vauto. Qed.

Lemma wstuck_lift:
  forall w sp s r r' ty n1 e1 n2 e2,
    wstuck (wval (wlift w) sp s r) r' ty n1 e1 n2 e2 <->
    wstuck (wval w (fst s) (snd s) r) r' ty n1 e1 n2 e2.
Proof. by split; intros; destruct H; simpls; des; try discriminate; vauto. Qed.

Lemma wstuck_disj:
  forall r r1 r2
    (gEQ : forall (ty : flextyp) n v n' v',
      r ty n v n' v' <-> r1 ty n v n' v' \/ r2 ty n v n' v')
    g ty n1 e1 n2 e2
    (EREL : wstuck r g ty n1 e1 n2 e2),
  wstuck r1 g ty n1 e1 n2 e2 \/ wstuck r2 g ty n1 e1 n2 e2.
Proof.
  intros; destruct EREL; clarify; eapply gEQ in FREL; des; vauto. 
Qed.
Implicit Arguments wstuck_disj [r r1 r2 g ty n1 e1 n2 e2].

Lemma wstuck_not_value1: forall r g ty n1 e1 n2 e2 (STK: wstuck r g ty n1 e1 n2 e2), 
  ~ is_value_of_exp e1.
Proof. i; depdes STK; subst; eauto. Qed.

Lemma wstuck_not_value2: forall r g ty n1 e1 n2 e2 (STK: wstuck r g ty n1 e1 n2 e2), 
  ~ is_value_of_exp e2.
Proof. i; depdes STK; subst; eauto. Qed.

Lemma wglob_prod1: forall 
    w1 w2 (WD: wdisj w1 w2)
    g (gOK: wglob (wlift (wprod w1 w2)) g) s2, 
  wglob (wlift w1) (fun s => g (fst s,(snd s, s2))).
Proof.
  intros; rdes gOK; intros; split; [|split]; unnw; intros; simpls; des; eauto.

  assert (X:=GSUP (fst s, (snd s,s2))); rdes X.
  repeat split; simpls; unnw; intros.
  - by apply LE; intuition. 
  - apply GEREF in R; simpls.
    assert (Z:= wlocal_nonref w2 (fst s) s2 (g (fst s,(snd s,s2))) ty n v n' v').
    by intuition. 
  - rdes NM.
    + by eapply GENAM in R; left; des; clarify; elimtype False; eauto using wnam_supports.
    + eapply GENAM in R; right; des; clarify; red; eauto. 
      by elimtype False; eauto using wnam_supports.
Qed.

Lemma wglob_prod2: forall 
    w1 w2 (WD: wdisj w1 w2)
    g (gOK: wglob (wlift (wprod w1 w2)) g) s1, 
    wglob (wlift w2) (fun s => g (fst s,(s1, snd s))).
Proof.
  intros; rdes gOK; intros; split; [|split]; unnw; intros; simpls; des; eauto.

  assert (X:= GSUP (fst s, (s1, snd s))); rdes X.
  repeat split; simpls; unnw; intros.
  - by apply LE; intuition. 
  - apply GEREF in R; simpls.
    assert (Z:= wlocal_nonref w1 (fst s) s1 (g (fst s,(s1, snd s))) ty n v n' v').
    by intuition.  
  - rdes NM.
    + by eapply GENAM in R; left; des; clarify; elimtype False; eauto using wnam_supports.
    + eapply GENAM in R; right; des; clarify; red; eauto.
      by elimtype False; eauto using wnam_supports.
Qed.

Lemma wexp_core_lreds: 
  forall w rr hf hf' hh hh' hh1 hh1' e e' e1 e1' s0 s s1 r g ty n n' n1 n1'
    (R: paco13 (wexp_core_gen w) rr r g s0 s1 ty hf hh1 n1 e1 hf' hh1' n1' e1')
    (RED : lreduces n hh e n1 hh1 e1)
    (RED' : lreduces n' hh' e' n1' hh1' e1')
    (EXT : wext w s1 s),
  paco13 (wexp_core_gen w) rr r g s0 s ty hf hh n e hf' hh' n' e'.
Proof.
  i; pdes R.
  - pfold; econs; eapply lreds_ldiv; eauto. 
  - pfold; eapply wexp_Terminate with (s':=s'); eauto using wext_trans, lreds_trans.
  - pfold; eapply wexp_Call with (s':=s'); eauto using wext_trans, lreds_trans.
Qed.

Lemma wexp_core_mon :
  forall w r r' g g' s0' s0 s ty hf hh n n1 e hf' hh' n' n1' e'
    (WE: wexp_core w r g s0' s ty hf hh n e hf' hh' n' e') 
    (LTR: r <6= r') (LTG: g <6= g') (PUB: wextpub w s0' s0) (LTN: n <= n1) (LTN': n' <= n1'), 
  wexp_core w r' g' s0 s ty hf hh n1 e hf' hh' n1' e'. 
Proof.
  intros until s0; revert r r' g g'.
  pcofix X with z; i; pdes WE. 
  - pfold; econs 1; eauto using ldiv_mon.
  - pfold; econs 2; eauto using wextpub_trans, vclos_mon, wheap_mon, lreds_mon.
  - pfold; econs 3; eauto using wheap_mon, wstuck_mon, lreds_mon.
    by ii; edestruct KREL; (inst; eauto).
Qed.

Lemma wexp_mon :
  forall w g s0' s0 s ty n1 n1' e1 n2 n2' e2 
    (WE: wexp w g s0 s ty n1 e1 n2 e2) 
    (PUB: wextpub w s0 s0')
    (LTN1: n1 <= n1') (LTN2: n2 <= n2'), 
  wexp w g s0' s ty n1' e1 n2' e2. 
Proof. ii; eapply wexp_core_mon; eauto. Qed.

Lemma wexp_u_mon: forall w wexp1 wexp2, 
  wexp1 <13= wexp2 -> wexp_u w wexp1 <9= wexp_u w wexp2.
Proof. ii; eapply H; eauto. Qed.

Lemma wcont_mon :
  forall w g g' s00 s00' s0 s0' s s' ty' ty n1 n1' K1 n2 n2' K2
    (REL: wcont w g s00 s0 s ty' ty n1 K1 n2 K2)
    (PUB00: wextpub w s00 s00')
    (PUB0: wextpub w s0' s0)
    (EXT: wext w s' s)
    (LEG: g <6= g')
    (LTN1: n1 <= n1') (LTN2: n2 <= n2'),
  wcont w g' s00' s0' s' ty' ty n1' K1 n2' K2.
Proof.
  ii; eapply wexp_mon; [eapply REL| | | | | |]
  ; eauto using wextpub_trans, wext_trans.
Qed.

Lemma wexp_val :
  forall w g (gOK: wglob w g) s0 s (PUB: wextpub w s s0) ty n (v: voe) n' (v': voe) 
    (V: vclos (g s) ty n (v_of v) n' (v_of v')),
    wexp w g s0 s ty n (e_of v) n' (e_of v'). 
Proof. 
  ii; destruct v, v'; pfold; ss; subst. 
  econs 2; eauto using lreduces_nil.
Qed.

Lemma wexp_app_nil :
  forall (w: world) g s0 s n n' ty e1 e2
    (CEQ: wexp w g s0 s ty n (app_cont [] e1) n' (app_cont [] e2)),
  wexp w g s0 s ty n e1 n' e2.
Proof. eauto. Qed.

Lemma wexp_core_app_cont :
  forall w g (gOK: wglob w g) s0 s0' s ty ty' nk K1 nk' K2 hf hh hf' hh' 
    (KREL: wcont w g s0 s0' s ty' ty nk K1 nk' K2)
    r (LT: g <6= r) 
    ne e ne' e' (EREL: wexp_core w r g s0' s ty' hf hh ne e hf' hh' ne' e')
    nkk (GUARD: is_value_of_exp e -> nkk >= nk)
    nkk' (GUARD': is_value_of_exp e' -> nkk' >= nk'),
  wexp_core w r g s0 s ty 
   hf hh (ne+nkk) (app_cont K1 e) hf' hh' (ne'+nkk') (app_cont K2 e').
Proof.
  pcofix CIH with z; intros; pdes EREL.
  - pfold; econs 1; eauto using diverges_cont, ldiv_div, div_ldiv.
  - exploit KREL; eauto; intros Y.
    eapply wexp_core_lreds.
    + by eapply wexp_core_mon in Y; [eapply paco13_mon| | | | |]; eauto.
    + destruct (classic(is_value_of_exp e)).
      * apply GUARD in H.
        eapply lreds_mon; [apply lreds_cont, lreds_plus| |]; eauto.
      * eauto 10 using lreds_cont, lreds_reds, reds_lreds_val, val_is_voe.
    + destruct (classic(is_value_of_exp e')).
      * apply GUARD' in H.
        eapply lreds_mon; [apply lreds_cont, lreds_plus| |]; eauto.
      * eauto 10 using lreds_cont, lreds_reds, reds_lreds_val, val_is_voe.
    + eauto.
  - apply lreds_cont with (K:=K1) in RED; rewrite <-app_cont_assoc in RED.
    apply lreds_cont with (K:=K2) in RED'; rewrite <-app_cont_assoc in RED'.
    pfold; econs 3; try eapply EREL; try eapply HREL; eauto using lreds_mon.
    ii; right; rewrite !app_cont_assoc.
    instantiate (1:= nk' + nk2); instantiate (1:= nk + nk1).
    replace (nk+nk1+nv) with (nv+nk1+nk) by omega.
    replace (nk'+nk2+nv') with (nv'+nk2+nk') by omega.
    eapply CIH; eauto using wcont_mon.
    eapply wexp_core_mon; [eapply or_false, KREL0| | | | |]; eauto.
Qed.

Lemma wexp_app_cont :
  forall w s0 s0' g (gOK: wglob w g) s ty ty' nk K1 nk' K2
    (KREL: wcont w g s0 s0' s ty' ty nk K1 nk' K2)
    ne e ne' e' (EREL: wexp w g s0' s ty' ne e ne' e')
    nkk (GUARD: is_value_of_exp e -> nkk >= nk)
    nkk' (GUARD': is_value_of_exp e' -> nkk' >= nk'),
  wexp w g s0 s ty (ne+nkk) (app_cont K1 e) (ne'+nkk') (app_cont K2 e').
Proof.
  by do 2 red; intros; eapply wexp_core_app_cont; eauto.
Qed.

Structure taggedExp := tExp0 { untagExp :> exp}.

Structure KESplit := mkKESplit 
  { term_of : taggedExp; scont_of : scont; body_of : exp; 
    KEsplit_pf : untagExp term_of = app_cont [scont_of] body_of }.

Lemma wexp_app_scont :
  forall (w: world) g s0 s n n' ty k1 ke1 k2 ke2
    (CEQ: wexp w g s0 s ty n (app_cont (scont_of ke1 :: k1) (body_of ke1))
                        n' (app_cont (scont_of ke2 :: k2) (body_of ke2))),
  wexp w g s0 s ty n (app_cont k1 (term_of ke1)) n' (app_cont k2 (term_of ke2)).
Proof. by intros; destructs ke1 ke2 term_of0 term_of1; subst; simpls; subst. Qed.

Definition tExp1  e := @tExp0  e.
Definition tExp2  e := @tExp1  e.
Definition tExp3  e := @tExp2  e.
Definition tExp4  e := @tExp3  e.
Definition tExp5  e := @tExp4  e.
Definition tExp6  e := @tExp5  e.
Definition tExp7  e := @tExp6  e.
Definition tExp8  e := @tExp7  e.
Definition tExp9  e := @tExp8  e.
Definition tExp10 e := @tExp9  e.
Definition tExp11 e := @tExp10 e.
Definition tExp12 e := @tExp11 e.
Definition tExp13 e := @tExp12 e.
Definition tExp14 e := @tExp13 e.
Definition tExp15 e := @tExp14 e.
Definition tExp16 e := @tExp15 e.
Definition tExp17 e := @tExp16 e.
Definition tExp18 e := @tExp17 e.
Definition tExp19 e := @tExp18 e.
Canonical Structure tExp20 e := @tExp19 e.

Canonical Structure KES_SKif ee e1 e2 :=
  @mkKESplit (tExp0 (Eif ee e1 e2)) (SKif e1 e2) ee refl_equal.
Program 
Canonical Structure KES_SKpairR ee ve :=
  @mkKESplit (tExp1 (Epair (e_of ve) ee)) (SKpairR (v_of ve)) ee _.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKpairL ee e :=
  @mkKESplit (tExp2 (Epair ee e)) (SKpairL e) ee refl_equal.
Canonical Structure KES_SKfst ee :=
  @mkKESplit (tExp3 (Efst ee)) SKfst ee refl_equal.
Canonical Structure KES_SKsnd ee :=
  @mkKESplit (tExp4 (Esnd ee)) SKsnd ee refl_equal.
Canonical Structure KES_SKinl ee :=
  @mkKESplit (tExp5 (Einl ee)) SKinl ee refl_equal.
Canonical Structure KES_SKinr ee :=
  @mkKESplit (tExp6 (Einr ee)) SKinr ee refl_equal.
Canonical Structure KES_SKcase ee x e1 e2 :=
  @mkKESplit (tExp7 (Ecase x ee e1 e2)) (SKcase x e1 e2) ee refl_equal.
Canonical Structure KES_SKroll ee :=
  @mkKESplit (tExp8 (Eroll ee)) SKroll ee refl_equal.
Canonical Structure KES_SKunroll ee :=
  @mkKESplit (tExp9 (Eunroll ee)) SKunroll ee refl_equal.
Program 
Canonical Structure KES_SKappR ee ve :=
  @mkKESplit (tExp10 (Eapp (e_of ve) ee)) (SKappR (v_of ve)) ee refl_equal.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKappL ee e :=
  @mkKESplit (tExp11 (Eapp ee e)) (SKappL e) ee refl_equal.
Canonical Structure KES_SKinst ee :=
  @mkKESplit (tExp12 (Einst ee)) SKinst ee refl_equal.
Canonical Structure KES_SKpack ee :=
  @mkKESplit (tExp13 (Epack ee)) SKpack ee refl_equal.
Canonical Structure KES_SKunpack ee x e :=
  @mkKESplit (tExp14 (Eunpack x ee e)) (SKunpack x e) ee refl_equal.
Canonical Structure KES_SKnew ee :=
  @mkKESplit (tExp15 (Enew ee)) SKnew ee refl_equal.
Canonical Structure KES_SKread ee :=
  @mkKESplit (tExp16 (Eread ee)) SKread ee refl_equal.
Program 
Canonical Structure KES_SKwriteR ee ve :=
  @mkKESplit (tExp17 (Ewrite (e_of ve) ee)) (SKwriteR (v_of ve)) ee refl_equal.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKwriteL ee e :=
  @mkKESplit (tExp18 (Ewrite ee e)) (SKwriteL e) ee refl_equal.
Program 
Canonical Structure KES_SKrefeqR ee ve :=
  @mkKESplit (tExp19 (Erefeq (e_of ve) ee)) (SKrefeqR (v_of ve)) ee refl_equal.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure KES_SKrefeqL ee e :=
  @mkKESplit (tExp20 (Erefeq ee e)) (SKrefeqL e) ee refl_equal.

(*----------------------------------------------------------------------------*)
(** ** External calls *)
(*----------------------------------------------------------------------------*)

Inductive cowexp_gen (w: world) r g hf hf' ty s0 cowexp : 
  wst w -> heap -> nat -> exp -> heap -> nat -> exp -> Prop :=
| cowexp_stop s hh n e hh' n' e'
    (OEQ: @wexp_core w r g s0 s ty hf hh n e hf' hh' n' e') :
  cowexp_gen w r g hf hf' ty s0 cowexp s hh n e hh' n' e'
| cowexp_steps s s1 hh hh' n e n' e' hh1 hh1' n1 e1 n1' e1' hh2 hh2' n2 e2 n2' e2'
    (REL: cowexp s1 hh2 n2 e2 hh2' n2' e2' : Prop)
    (EXT: wext w s1 s) 
    (STEP1: lreduces n hh e n1 hh1 e1) (STEPS1: lreduce n1 hh1 e1 n2 hh2 e2)
    (STEP2: lreduces n' hh' e' n1' hh1' e1') (STEPS2: lreduce n1' hh1' e1' n2' hh2' e2'):
  cowexp_gen w r g hf hf' ty s0 cowexp s hh n e hh' n' e'
.
Hint Constructors cowexp_gen.

Definition cowexp w r g hf hf' ty s0 s hh n e hh' n' e' := 
  paco7 (cowexp_gen w r g hf hf' ty s0) bot7 s hh n e hh' n' e'.
Lemma cowexp_gen_mon: 
  forall w r g hf hf' ty s0, monotone7 (cowexp_gen w r g hf hf' ty s0). 
Proof. pmonauto. Qed.
Hint Unfold cowexp.
Hint Resolve cowexp_gen_mon : paco.

Lemma cowexp_div_or_wexp: forall w r g hf hf' ty s0 n n' s hh e hh' e'
    (REL: cowexp w r g hf hf' ty s0 s hh n e hh' n' e'),
  (ldiverges n hh e /\ ldiverges n' hh' e') \/
  (wexp_core w r g s0 s ty hf hh n e hf' hh' n' e').
Proof.
  i; destruct (classic(wexp_core w r g s0 s ty hf hh n e hf' hh' n' e')); eauto.
  r in H; left; split;
  ( revert_until s0; pcofix CIH with z 
  ; i; punfold REL; depdes REL; [by exfalso; eauto|]
  ; pclearbot; eapply lreds_ldiv; eauto 15 using wexp_core_lreds, lreds_trans, lreduces_nil).
Qed.

Lemma wexp_core_external :
  forall w g r s0 s ty hf hf' hh hh' n e n' e' 
    (CONS: wconsistent w) 
    (gOK: wglob w g)
    (gEQ: forall s n n' ty v v', g s ty n v n' v' 
             <-> wval w () s (g s) ty n v n' v' \/ r s ty n v n' v')
    (REL: wexp_core w g g s0 s ty hf hh n e hf' hh' n' e'),
  wexp_core w r g s0 s ty hf hh n e hf' hh' n' e'.
Proof.
  i; cut (cowexp w r g hf hf' ty s0 s hh n e hh' n' e').
    by intro X; eapply cowexp_div_or_wexp in X; des; eauto.
  revert s hh n e hh' n' e' REL; pcofix CIH with z.
  i; pfold; pdes REL; eauto.
  eapply wstuck_disj in EREL; [|eauto]; des; [|econs; eauto].
  edestruct CONS; eauto; des.
  assert (LRED1: exists nn1, (forall h, lreduce ne1 h e1 (n1+nn1) h e0)
                 /\ (is_value_of_exp e0 -> nn1 >= nk1)).
  { destruct (classic(is_value_of_exp e0)).
    - eexists nk1; split; eauto.
      i; specialize (LRED h); depdes LRED; eauto.
      by exfalso; eauto using wstuck_not_value1.
    - eexists 0; rewrite plus_0_r; split; eauto; done. }
  assert (LRED2: exists nn2, (forall h, lreduce ne2 h e2 (n1'+nn2) h e1')
                  /\ (is_value_of_exp e1' -> nn2 >= nk2)).
  { destruct (classic(is_value_of_exp e1')).
    - eexists nk2; split; eauto.
      i; specialize (LRED' h); depdes LRED'; eauto.
      by exfalso; eauto using wstuck_not_value2.
    - eexists 0; rewrite plus_0_r; split; eauto; done. }
  clear LRED LRED'; des.
  econs 2; [|M|M|M|M|M]; Mdo by eauto using lred_cont.
  right; apply CIH; eapply mp; [eapply (CONS _ gOK); eauto|].
  intro X; eapply wexp_core_app_cont; eauto.
  ii; eapply or_false, KREL; eauto.
Qed.

Lemma wexpgen_external :
  forall w (CONS: wconsistent w) g r s0 s ty n e n' e' 
    (gOK: wglob w g) 
    (gEQ: forall s ty n v n' v', g s ty n v n' v' 
             <-> wval w () s (g s) ty n v n' v' \/ r s ty n v n' v') 
    (REL: wexp w g s0 s ty n e n' e'),
  wexpgen w r g s0 s ty n e n' e'.
Proof. by repeat intro; eapply wexp_core_external; eauto. Qed.

Corollary wexp_no_external :
  forall w (CONS: wconsistent w) 
    g (gOK: wglob w g)  
    (gEQ: forall s ty n v n' v', g s ty n v n' v' <-> wval w () s (g s) ty n v n' v')
    ty n e n' e' s0 s (H: wexp w g s0 s ty n e n' e'), 
  wexpgen w bot6 g s0 s ty n e n' e'.
Proof.
  intros; eapply wexpgen_external; eauto.
  intros; specialize (gEQ s1 ty0 n0 v n'0 v'); intuition.
Qed.
