Require Import basis lang subs countable.
Require Import vgtac Program List Relations.
Require Export paco.

Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** * Basic library definitions *)
(*----------------------------------------------------------------------------*)

(* begin hide *)
Obligation Tactic := 
  repeat intro; 
  try (eby solve [unnw; program_simpl]); 
  try (eby simpls; des; des; unnw; clarify; eauto 10).
(* end hide *)

(*----------------------------------------------------------------------------*)
(** ** Divergence and stuckness *)
(*----------------------------------------------------------------------------*)

Inductive lreduce n h e : nat -> heap -> exp -> Prop :=
| lred_log: forall n' (LT: n' < n), lreduce n h e n' h e
| lred_phy: forall n' h' e' (RED: reduce h e h' e'), lreduce n h e n' h' e'
.
Hint Constructors lreduce.

Inductive lreduces n h e : nat -> heap -> exp -> Prop :=
| lreduces_nil : lreduces n h e n h e
| lreduces_step : forall n1 h1 e1 n2 h2 e2
    (STEP: lreduce n h e n1 h1 e1)
    (RED : lreduces n1 h1 e1 n2 h2 e2),
  lreduces n h e n2 h2 e2.
Hint Resolve lreduces_step.

Inductive ldiverges_gen diverges n h e : Prop :=
| ldiverges_step : forall n' h' e' 
    (STEP: lreduce n h e n' h' e')
    (DIV: diverges n' h' e': Prop), 
  ldiverges_gen diverges n h e.
Hint Constructors ldiverges_gen.

Definition ldiverges n h e := paco3 ldiverges_gen bot3 n h e.
Hint Unfold ldiverges.
Lemma ldiverges_gen_mon: monotone3 ldiverges_gen. Proof. pmonauto. Qed.
Hint Resolve ldiverges_gen_mon : paco.

(*----------------------------------------------------------------------------*)
(** * Various relations *)
(*----------------------------------------------------------------------------*)

Implicit Types (v : val) (e: exp) (h : heap).

(** Flexible types are those for which value equivalence is not fixed. *)

Inductive flextyp : Type := 
  | FTarrow (ty1 ty2 : typ)
  | FTref (ty : typ)
  | FTforall (ty : typ)
  | FTname (a : typename).

Definition unflex (ty : flextyp) : typ :=
  match ty with
  | FTarrow ty1 ty2 => Tarrow ty1 ty2
  | FTref ty' => Tref ty'
  | FTforall ty' => Tforall ty'
  | FTname a => Tname a
  end.
Coercion unflex : flextyp >-> typ.

Lemma unflex_inj : forall t1 t2, unflex t1 = unflex t2 -> t1 = t2.
Proof.
  by destruct t1; intros; destruct t2; inversion H; try rewrite <- H1, <- H2.
Qed.


(** Here are the basic domains that we will be using: *)

Definition vrelf :=flextyp -> nat -> val -> nat -> val -> Prop.
Definition vrel := typ -> nat -> val -> nat -> val -> Prop.
Definition erel := typ -> nat -> exp -> nat -> exp -> Prop.
Definition hrel := heap -> heap -> Prop.

Hint Unfold vrelf vrel erel hrel.

Definition vrelf_closed (r : vrelf) := 
  forall ty n v n' v' (V: r ty n v n' v'),
  <<CV: fv_exp v = [] >> /\ <<CV': fv_exp v' = [] >>.

Hint Unfold vrelf_closed.

(*----------------------------------------------------------------------------*)
(** * Value Closure *)
(*----------------------------------------------------------------------------*)

(** The following definition, [vclos], lifts the definition of value 
    equivalence from flexible types only to all types. *)

Inductive vclos_gen (r: vrelf) (vr: vrel) : vrel :=
  | vclos_gen_unit : forall n n', vclos_gen r vr Tunit n Vunit n' Vunit
  | vclos_gen_int  : forall n n' m, vclos_gen r vr Tint n (Vint m) n' (Vint m)
  | vclos_gen_bool_true : forall n n', vclos_gen r vr Tbool n Vtrue n' Vtrue
  | vclos_gen_bool_false : forall n n', vclos_gen r vr Tbool n Vfalse n' Vfalse
  | vclos_gen_prod:
    forall ty1 n1 v1 n1' v1' (VR1: vr ty1 n1 v1 n1' v1') 
           ty2 n2 v2 n2' v2' (VR2: vr ty2 n2 v2 n2' v2')
           n n',
      vclos_gen r vr (Tprod ty1 ty2) n (Vpair v1 v2) n' (Vpair v1' v2')
  | vclos_gen_sum_inl:
    forall ty1 n1 v1 n1' v1' (VR: vr ty1 n1 v1 n1' v1') ty2 n n',
      vclos_gen r vr (Tsum ty1 ty2) n (Vinl v1) n' (Vinl v1')
  | vclos_gen_sum_inr:
    forall ty2 n2 v2 n2' v2' (VR: vr ty2 n2 v2 n2' v2') ty1 n n',
      vclos_gen r vr (Tsum ty1 ty2) n (Vinr v2) n' (Vinr v2')
  | vclos_gen_mu:
    forall ty n1 v n1' v' (VR: vr (tsubst1 0 (Tmu ty) ty) n1 v n1' v') n n',
      vclos_gen r vr (Tmu ty) n (Vroll v) n' (Vroll v')
  | vclos_gen_exists:
    forall ty' (WFty': lc_typ 0 ty') ty n1 v n1' v' (VR: vr (tsubst1 0 ty' ty) n1 v n1' v') n n',
      vclos_gen r vr (Texists ty) n (Vpack v) n' (Vpack v')
  | vclos_gen_name : 
    forall aa n v n' v' (VR: r (FTname aa) n v n' v'), vclos_gen r vr (Tname aa) n v n' v'
  | vclos_gen_arrow:
    forall ty1 ty2 n v n' v' (VR: r (FTarrow ty1 ty2) n v n' v'), vclos_gen r vr (Tarrow ty1 ty2) n v n' v'
  | vclos_gen_forall:
    forall ty n v n' v' (VR: r (FTforall ty) n v n' v'), vclos_gen r vr (Tforall ty) n v n' v'
  | vclos_gen_ref:
    forall ty n1 v n1' v' (VR: r (FTref ty) n1 v n1' v') n n', vclos_gen r vr (Tref ty) n v n' v'.
Hint Constructors vclos_gen.

Inductive vclos (r : vrelf) (ty: typ) n1 (v1: val) n2 (v2: val) : Prop :=
| vclos_intro (REL: vclos_gen r (vclos r) ty n1 v1 n2 v2).
Hint Constructors vclos.

Lemma vclos_gen_mon : forall r0 r ty n1 v1 n2 v2 (REL: vclos_gen r0 r ty n1 v1 n2 v2) r' (LE: r <5= r'),
  vclos_gen r0 r' ty n1 v1 n2 v2.
Proof. intros; dependent destruction ty; inversion REL; eauto. Qed.

Lemma vclos_induction : 
  forall 
    r ty n1 v1 n2 v2 (EL: vclos r ty n1 v1 n2 v2)
    P (PF: forall ty n1 v1 n2 v2 (IH: vclos_gen r P ty n1 v1 n2 v2), P ty n1 v1 n2 v2),
  P ty n1 v1 n2 v2.
Proof.
  fix 7.
  intros; apply PF; destruct EL; destruct REL;
  econstructor; eauto; eapply vclos_induction; eauto.
Qed.

Lemma vclos_induction_strong : 
  forall 
    r ty n1 v1 n2 v2 (EL: vclos r ty n1 v1 n2 v2)
    P (PF: forall ty n1 v1 n2 v2 
             (IH: vclos_gen r (fun ty' n1' v1' n2' v2' => P ty' n1' v1' n2' v2' /\ vclos r ty' n1' v1' n2' v2') ty n1 v1 n2 v2), 
           P ty n1 v1 n2 v2),
  P ty n1 v1 n2 v2.
Proof.
  intros; cut (P ty n1 v1 n2 v2 /\ vclos r ty n1 v1 n2 v2); [intuition|].
  revert P PF; eapply (vclos_induction EL); clear ty v1 v2 EL; intros.
  split; eauto using vclos_gen_mon.
  econstructor; eapply vclos_gen_mon; eauto.
  by simpl; intros; hdes; eauto.
Qed.

Lemma vclos_mon: 
  forall g g' ty n v n' v' (VC: vclos g ty n v n' v') (LT: g <5= g'), 
    vclos g' ty n v n' v'.
Proof. 
  intros until v'; intros REL; eapply (vclos_induction REL); intros.
  by inv IH; clear IH; (econs; econs; eauto).
Qed.

Lemma vclos_mon_lv: forall g ty n1 v1 n2 v2
    (REL: vclos g ty n1 v1 n2 v2)
    (MONG: forall ty n1 n1' n2 n2' 
             (REL: g ty n1 v1 n2 v2) (LEN1: n1 <= n1') (LEN2: n2 <= n2'), 
           g ty n1' v1 n2' v2)
    n1' (LEN1: n1 <= n1') n2' (LEN2: n2 <= n2'),
  vclos g ty n1' v1 n2' v2.
Proof. induction 1; destruct REL; i; eauto. Qed.

Lemma vclos_closed : 
  forall r ty n v n' v' (VREL : vclos r ty n v n' v') (C : vrelf_closed r), 
    <<CV: fv_exp v = [] >> /\ <<CV': fv_exp v' = [] >>.
Proof.
  intros; pattern ty, n, v, n', v'; eapply (vclos_induction VREL); i.
  destruct IH; simpl; subst; try specialize (C _ _ _ VR); des;
    simpls; repeat (match goal with [H: _ = [] |- _] => rewrite H end);
    eauto 10 using lc_tsubst1D, ftn_tsubst1_mon_closed.
Qed.

(*----------------------------------------------------------------------------*)
(** * Dependent World *)
(*----------------------------------------------------------------------------*)

Structure world_dep (pst : Type) (pext : relation pst) := 
  { wnam : typename -> Prop 
  ; wst  : Type
  ; wext : relation wst
  ; wextpub : relation wst
  ; wval  : pst -> wst -> vrelf -> vrelf
  ; wheap : pst -> wst -> vrelf -> hrel
  ; wext_trans      : forall x y (EXT: wext x y) z (EXT': wext y z), wext x z 
  ; wextpub_refl    : forall x, wextpub x x 
  ; wextpub_trans   : forall x y (EXT: wextpub x y) z (EXT': wextpub y z), wextpub x z 
  ; wextpub_in_wext : forall x y (EXTP: wextpub x y), wext x y
  ; wval_mon        : forall sp s r ty n v n' v' (WV: wval sp s r ty n v n' v')
                        sp' (RP: pext sp' sp) 
                        s'  (R: wext s' s) 
                        r'  (LT: r <5= r')
                        n1  (LN: n <= n1)
                        n1' (LN': n' <= n1'),
                      wval sp' s' r' ty n1 v n1' v'
  ; wval_closed     : forall sp s r (C: vrelf_closed r), vrelf_closed (wval sp s r)
  ; wheap_mon       : forall sp s r h h' (PR: wheap sp s r h h') r' (LT: r <5= r'),
                        wheap sp s r' h h'
  ; wnam_supports   : forall sp s r a n v n' v' (WV: wval sp s r (FTname a) n v n' v'), wnam a 
  }.

Hint Immediate wextpub_refl wextpub_in_wext.

Lemma wext_refl pst (pext : relation pst) (w : world_dep pext) x : wext w x x.
Proof. apply wextpub_in_wext, wextpub_refl. Qed.
Hint Immediate wext_refl.

(*----------------------------------------------------------------------------*)
(** * Full World *)
(*----------------------------------------------------------------------------*)

Definition world := world_dep (fun x y: unit => True).

(*----------------------------------------------------------------------------*)
(** * Ref World *)
(*----------------------------------------------------------------------------*)

(** We define a ref world that defines the meaning of the
    mutable reference type. *)

Definition wrefst (s : typ -> location -> location -> Prop) :=
     (forall ty loc1 loc2 (IN : s ty loc1 loc2), <<LC: lc_typ 0 ty >>)
  /\ (forall ty loc1 loc2 (IN : s ty loc1 loc2) ty' loc1' loc2' (IN' : s ty' loc1' loc2'),
        (forall (EQ: loc1 = loc1'), <<TEQ: ty = ty' >> /\ <<LEQ: loc2 = loc2' >>) /\
        (forall (EQ: loc2 = loc2'), <<TEQ: ty = ty' >> /\ <<LEQ: loc1 = loc1' >>))
  /\ (exists tydom ldom, 
       forall ty l1 l2 (IN: s ty l1 l2), 
         <<INTY: In ty tydom>> /\ <<INL1: In l1 ldom>> /\ <<INL2: In l2 ldom>>).

Program Definition wref : world :=
  {| wnam n := False
   ; wst := { s : typ -> location -> location -> Prop | wrefst s } 
   ; wext s s' := s' <3= s
   ; wextpub s s' := s' <3= s
   ; wval u s r ty n v n' v' := exists loc, exists loc', exists ty',
                              <<NLT: n > 0>> /\ <<NLT': n' > 0>>
                           /\ <<TEQ: ty = FTref ty'>> 
                           /\ <<VEQ: v = Vloc loc >> 
                           /\ <<VEQ': v' = Vloc loc'>> 
                           /\ <<TLL: s ty' loc loc'>>
   ; wheap u s r h h' := 
       <<HSAT : forall ty loc loc' (R: s ty loc loc'), 
                exists n v n' v', 
                <<HL: h loc = Some v>> /\ <<HL': h' loc' = Some v'>> /\
                <<VC: vclos r ty n v n' v'>> >> /\
       <<HDOM1 : forall l v (IN: h l = Some v), exists ty l', s ty l l' >> /\
       <<HDOM2 : forall l' v' (IN: h' l' = Some v'), exists ty l, s ty l l' >> |}.
Next Obligation.
des; unnw; esplits; eauto; omega.
Qed.
Next Obligation.
des; unnw; esplits; ii; eauto; eapply HSAT in R; des; eauto 10 using vclos_mon.
Qed.

Program Definition sref0 : wst wref := exist _ (fun _ _ _ => False) _.
Next Obligation.
red; repeat (split; intros; vauto).
Existential 1 := nil. Existential 1 := nil.
Qed.

(*----------------------------------------------------------------------------*)
(** * Local World *)
(*----------------------------------------------------------------------------*)

Structure world_local := 
  { wldlocal :> world_dep (wext wref)
  ; wlocal_nonref : forall sp s r ty n v n' v', ~ wval wldlocal sp s r (FTref ty) n v n' v'
  }.

(*----------------------------------------------------------------------------*)
(** * Product of two local worlds *)
(*----------------------------------------------------------------------------*)

Program Definition wprod (w1 w2 : world_local) : world_local :=
  {| wldlocal := 
  {| wnam := nunion (wnam w1) (wnam w2)
   ; wext s s' := <<WE: wext w1 (fst s) (fst s')>> /\ <<WE': wext w2 (snd s) (snd s')>>
   ; wextpub s s' := <<WEP: wextpub w1 (fst s) (fst s')>> /\ <<WEP': wextpub w2 (snd s) (snd s')>>
   ; wval sp s r ty n v n' v' := wval w1 sp (fst s) r ty n v n' v' \/ wval w2 sp (snd s) r ty n v n' v'
   ; wheap sp s r h h' := exists h1 h1' h2 h2', 
                          <<WH:  wheap w1 sp (fst s) r h1 h1'>> /\
                          <<WH': wheap w2 sp (snd s) r h2 h2'>> /\
                          <<HP: hplus h1 h2 h>> /\ <<HP': hplus h1' h2' h'>> |} |}.
Next Obligation. simpls; des; eauto 10 using wext_trans, wextpub_trans. Qed. 
Next Obligation. simpls; des; eauto 10 using wext_trans, wextpub_trans. Qed.
Next Obligation. simpls; des; eauto 10 using wval_mon. Qed.
Next Obligation. des; eapply wval_closed; eauto. Qed.
Next Obligation. eby des; repeat eexists; try eapply wheap_mon. Qed.
Next Obligation. by des; constructor (eby eapply wnam_supports). Qed.
Next Obligation. simpls; des; eauto using wlocal_nonref. Qed.

(*----------------------------------------------------------------------------*)
(** * Lifting of a local world *)
(*----------------------------------------------------------------------------*)

Program Definition wlift (w : world_local) : world :=
{| wnam := nunion (wnam wref) (wnam w)
 ; wext s s' := <<WRE: wext wref (fst s) (fst s')>> /\ <<WRE': wext w (snd s) (snd s')>>
 ; wextpub s s' := <<WREP: wextpub wref (fst s) (fst s')>> /\ <<WREP': wextpub w (snd s) (snd s')>>
 ; wval sp s r ty n v n' v' := wval wref () (fst s) r ty n v n' v' \/ wval w (fst s) (snd s) r ty n v n' v'
 ; wheap sp s r h h' := exists h1 h1' h2 h2', 
                        <<WH: wheap wref () (fst s) r h1 h1'>> /\
                        <<WH': wheap w (fst s) (snd s) r h2 h2'>> /\
                        <<HP: hplus h1 h2 h>> /\ <<HP': hplus h1' h2' h'>> |}.
Next Obligation. des; eauto 10 using wext_trans. Qed.
Next Obligation. des; eauto 10 using wextpub_trans. Qed.
Next Obligation. des; eauto 10 using wval_mon. Qed.
Next Obligation. des; eapply wval_closed; eauto. Qed.
Next Obligation. 
des; simpls; do 4 eexists; repeat split; try eapply wheap_mon; des; unnw; eauto.
by intros; specialize (HSAT _ _ _ R); des; eauto 10 using vclos_mon.
Qed.
Next Obligation. by des; constructor (eby eapply wnam_supports). Qed.

(*----------------------------------------------------------------------------*)
(** * Single State Worlds *)
(*----------------------------------------------------------------------------*)

(* bug in Coq8.4: will be fixed in the next version *)

(*
Program Definition wsingle (VR: vrelf->vrelf ) (HR: vrelf->hrel) 
  (VRMON: forall r r' (LE: r <4= r'), VR r <4= VR r')
  (HRMON: forall r r' (LE: r <4= r'), HR r <2= HR r')
  (VRCLOS: forall r (C: vrelf_closed r), vrelf_closed (VR r))
  (WFARR: forall r ty' ty f1 f2, ~ VR r 0 (FTarrow ty' ty) f1 f2)
  (WFINST: forall r ty f1 f2, ~ VR r 0 (FTforall ty) f1 f2)
  : world_local 
  :=
  {| wldlocal :=
  {| wnam := fun _ => False
   ; wst := unit
   ; wext := fun _ _ => True
   ; wextpub := fun _ _ => True
   ; wval := fun _ _ r n ty v v' => 
     match ty with | FTref _ => False | FTname _ => False | _ => VR r n ty v v' end
   ; wheap := fun _ _ => HR |} |}.
Next Obligation. destruct ty; simpls; eapply VRMON; eauto. Qed.
Next Obligation. destruct ty; try done; eapply VRCLOS; eauto. Qed.
*)

Definition wsingle (vr: vrelf->vrelf ) (hr: vrelf->hrel) 
  (VRMON: forall r ty n1 v1 n2 v2 
            (VREL: vr r ty n1 v1 n2 v2) r' 
            (LER: r <5= r') n1' (LEN1: n1 <= n1') n2' (LEN2: n2 <= n2'), 
          vr r' ty n1' v1 n2' v2)
  (HRMON: forall r r' (LE: r <5= r'), hr r <2= hr r')
  (VRCLOS: forall r (C: vrelf_closed r), vrelf_closed (vr r))
  (WFVAL: forall r ty n1 v1 n2 v2, vr r ty n1 v1 n2 v2 -> n1>0 /\ n2>0)
  : world_local.

refine(
  {| wldlocal :=
  {| wnam := fun _ => False
   ; wst := unit
   ; wext := fun _ _ => True
   ; wextpub := fun _ _ => True
   ; wval := fun _ _ r ty n v n' v' => 
     match ty with | FTref _ => False | FTname _ => False | _ => vr r ty n v n' v' end
   ; wheap := fun _ _ => hr |} |}).

eauto.
Grab Existential Variables.
- eauto. 
- s; eauto.
- by intros; destruct ty; try done; eapply VRCLOS; eauto.
- by intros; destruct ty; simpls; eapply VRMON; eauto.
- eauto. 
- eauto. 
- unnw; eauto. 
- eauto.
Defined.

(*----------------------------------------------------------------------------*)
(** * Global Knowledge *)
(*----------------------------------------------------------------------------*)

(** A custom order between two value relations that requires the
    meaning of reference and abstract types to agree between the
    two relations. *)

Definition lerefn N (r r' : vrelf) :=
  <<LE: forall ty n v n' v' (R: r ty n v n' v'), r' ty n v n' v' >> /\
  <<GEREF: forall ty n v n' v' (R: r' (FTref ty) n v n' v'), r (FTref ty) n v n' v' >> /\
  <<GENAM: forall a n v n' v' (NM: N a) (R: r' (FTname a) n v n' v'), r (FTname a) n v n' v' >>.

(** A global value relation [g] is consistent w.r.t. a world [w] if
    it is (i) well-formed; 
        (ii) monotonic w.r.t state transitions 
    and (iii) relates any two values that are related by [w]. *)

Definition wglob (w : world) (g : wst w -> vrelf) :=
  <<GCLO: forall s, vrelf_closed (g s) >> /\
  <<GMON: forall s s' ty n n1 v n' n1' v' 
            (R: g s ty n v n' v') 
            (EXT: wext w s' s) (LEN: n <= n1) (LEN': n' <= n1'), 
          g s' ty n1 v n1' v' >> /\
  <<GSUP: forall s, lerefn (wnam w) (wval w () s (g s)) (g s) >>.

(*----------------------------------------------------------------------------*)
(** * Expression and Continuation Equivalence *)
(*----------------------------------------------------------------------------*)

(** 
   Co-inductive definition of when two (closed) expressions are related. 
*)

Inductive wstuck (Rf : vrelf) (Rv : vrelf) ty n1 e1 n2 e2 : Prop :=
  | wstuck_App
      ty' nv1 nv2 (f1 v1 f2 v2 : val)
      (EQ1  : e1 = Eapp f1 v1)
      (EQ2  : e2 = Eapp f2 v2)
      (FREL : Rf (FTarrow ty' ty) n1 f1 n2 f2)
      (VREL : vclos Rv ty' nv1 v1 nv2 v2)
  | wstuck_Inst
      ty' ty0 (f1 f2 : val)
      (EQ1  : e1 = Einst f1)
      (EQ2  : e2 = Einst f2)
      (EQ3  : ty = tsubst1 0 ty' ty0)
      (LC   : lc_typ 0 ty')
      (FREL : Rf (FTforall ty0) n1 f1 n2 f2).
Hint Constructors wstuck.

Definition wexp_u (w: world) wexp_core (r g: wst w -> vrelf) (s0 s: wst w) (ty: typ) (n:nat) e (n':nat) e' :=
  forall 
    h h' (HREL: wheap w () s (g s) h h')
    hf hh (HPLUS: hplus h hf hh)
    hf' hh' (HPLUS': hplus h' hf' hh'),
    wexp_core r g s0 s ty hf hh n e hf' hh' n' e' : Prop.

Definition wcont_u w wexp_core g s00 s0 s ty' ty nk k nk' k' := 
  forall s' (EXTP: wextpub w s' s0) (EXT: wext w s' s) g' (g'OK: wglob w g') (LT : g <6= g')
         nv v nv' v' (VREL: vclos (g' s') ty' nv v nv' v'),
  wexp_u w wexp_core g' g' s00 s' ty (nk+nv) (app_cont k v) (nk'+nv') (app_cont k' v').

Inductive wexp_core_gen (w : world) wexp_core
  (r g: wst w -> vrelf) (s0 s: wst w) (ty: typ) (hf hh: heap) (n: nat) (e:exp) (hf' hh': heap) (n': nat) (e': exp) : Prop :=
| wexp_Diverge
    (DIV: ldiverges n hh e) 
    (DIV': ldiverges n' hh' e')

| wexp_Terminate
     s' (EXT: wext w s' s) (EXTP: wextpub w s' s0)
     nv1 nv2 (v1 v2 : val) (VREL: vclos (g s') ty nv1 v1 nv2 v2)
     h1 h2 (HREL: wheap w () s' (g s') h1 h2)
     hh1 (PLUS1: hplus h1 hf hh1)
     hh2 (PLUS2: hplus h2 hf' hh2)
     (RED: lreduces n hh e nv1 hh1 v1)
     (RED': lreduces n' hh' e' nv2 hh2 v2)

| wexp_Call
     s' (EXT: wext w s' s)
     h1 h2 (HREL: wheap w () s' (g s') h1 h2)
     e1 e2 ne1 ne2 ty' (EREL: wstuck (r s') (g s') ty' ne1 e1 ne2 e2)
     nk1 nk2 K1 K2 (KREL: wcont_u w wexp_core g s0 s' s' ty' ty nk1 K1 nk2 K2)
     hh1 (PLUS1: hplus h1 hf hh1)
     hh2 (PLUS2: hplus h2 hf' hh2)
     (RED: lreduces n hh e ne1 hh1 (app_cont K1 e1))
     (RED': lreduces n' hh' e' ne2 hh2 (app_cont K2 e2))
.
Hint Constructors wexp_core_gen.

Definition wexp_core w r g s0 s ty hf hh n e hf' hh' n' e':= 
  paco13 (wexp_core_gen w) bot13 r g s0 s ty hf hh n e hf' hh' n' e'.
Hint Unfold wexp_core.
Lemma wexp_core_gen_mon: forall w, monotone13 (wexp_core_gen w).
Proof.
  ii; destruct IN; [econs 1|econs 2|econs 3]; eauto.
  ii; eapply LE, KREL; eauto.
Qed.
Hint Resolve wexp_core_gen_mon : paco.

Definition wexp w g := wexp_u w (wexp_core w) g g.

Definition wexpgen w r g := wexp_u w (wexp_core w) r g.

(** Two continuations are related if whenever we apply them to 
    related arguments, we get related expressions. *)

Definition wcont w := wcont_u w (wexp_core w).

(*----------------------------------------------------------------------------*)
(** * Program Equivalence *)
(*----------------------------------------------------------------------------*)

Definition wstable (w : world_local) :=
  forall 
    g (gOK: wglob (wlift w) g)
    sp s h1 h2 (HR: wheap w sp s (g (sp,s)) h1 h2)
    sp' (EXT: wext _ sp' sp)
    hs1 hs2 (SHR: wheap wref () sp' (g (sp',s)) hs1 hs2)
    hh1 (DIS: hplus hs1 h1 hh1)
    hh2 (DIS: hplus hs2 h2 hh2),
  exists s', <<WEP: wextpub _ s' s>> /\ <<WH: wheap _ sp' s' (g (sp',s')) h1 h2>>.

(** We say that a world is inhabited iff it relates the empty heap to itself. *)

Definition winhabited (w : world) := 
  forall g (gOK: wglob w g), exists s0, wheap w () s0 (g s0) hemp hemp.

(** A world is consistent if it any functions it declares as 
    being equivalent are indeed so when applied to equivalent
    arguments. *)

Inductive drop (r : erel) ty n e n' e' : Prop :=
_drop n1 e1 n1' e1'
  (R: r ty n1 e1 n1' e1')
  (LRED: forall h, lreduce n h e n1 h e1)
  (LRED': forall h, lreduce n' h e' n1' h e1')
.

Definition wconsistent (w : world) :=
  forall g (gOK: wglob w g) s,
  wstuck (wval w () s (g s)) (g s) <5= drop (wexp w g s s).

(** Two expressions are related by a world [w] in a typing environment [D; G],
    if they are related by [w] after substituting any pairwise related values
    for the expressions' free variables. *)

Inductive wequiv D G w ty nf1 e1 nf2 e2 : Prop :=
  wequiv_intro
    (envOK: forall x ty0 (IN: In (x,ty0) G), lc_typ 0 ty0)
    (tyOK: lc_typ 0 ty)
    (FV1 : list_minus eq_termvar (fv_exp e1) (map fst G) = [])
    (FV2 : list_minus eq_termvar (fv_exp e2) (map fst G) = [])
    (WEXP: forall s
             g (gOK: wglob w g)
             d (DEQ: map fst d = D)
               (dOK: forall x (IN: In x d), lc_typ 0 (snd x))
             gg (GEQ: map fst gg = G) 
                (ggOK: forall x ty0 nv1 v1 nv2 v2 (IN: In ((x,ty0),((nv1,v1),(nv2,v2))) gg), 
                       vclos (g s) (tsubst_typ d ty0) nv1 v1 nv2 v2),
           wexp w g s s (tsubst_typ d ty) 
             (nf1 (map (fun x=>fst(fst(snd x))) gg)) 
             (subst_exp (map (fun x=>(fst(fst x),exp_of_val(snd(fst(snd x))))) gg) e1)
             (nf2 (map (fun x=>fst(snd(snd x))) gg)) 
             (subst_exp (map_pair fst (fun x=>exp_of_val(snd(snd x))) gg) e2)).

Hint Constructors wequiv.

Definition weq D G w ty nf1 e1 nf2 e2 : Prop :=
  <<INH : winhabited w>> /\
  <<CS : wconsistent w>> /\
  <<EQ : wequiv D G w ty nf1 e1 nf2 e2>>.
Hint Unfold weq.

(** Finally, two expressions are equivalent at a given type, if there 
    exists an inhabited, non-ref world, whose product with the global
    [wref] world is consistent and relates the two expressions. *)

Definition wreq (D : list typename) (G: list (termvar * typ)) ty e1 e2 :=
  forall N (INF: countably_infinite N), 
  exists w : world_local, exists nf1 nf2,
    <<NAM: wnam w <1= N >> /\
    <<WSTA: wstable w >> /\
    <<WEQ: weq D G (wlift w) ty nf1 e1 nf2 e2>>.

