Require Import basis lang subs countable.
Require Import vgtac Program List Relations.
Require Export paco.

Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** * Basic library definitions *)
(*----------------------------------------------------------------------------*)

(* begin hide *)
Obligation Tactic := 
  repeat intro; 
  try (eby solve [unnw; program_simpl]); 
  try (eby simpls; des; des; unnw; clarify; eauto 10).
(* end hide *)

(*----------------------------------------------------------------------------*)
(** ** Divergence and stuckness *)
(*----------------------------------------------------------------------------*)

Inductive lreduce n h e : nat -> heap -> exp -> Prop :=
| lred_log: forall n' (LT: n' < n), lreduce n h e n' h e
| lred_phy: forall n' h' e' (RED: reduce h e h' e'), lreduce n h e n' h' e'
.
Hint Constructors lreduce.

Inductive lreduces n h e : nat -> heap -> exp -> Prop :=
| lreduces_nil : lreduces n h e n h e
| lreduces_step : forall n1 h1 e1 n2 h2 e2
    (STEP: lreduce n h e n1 h1 e1)
    (RED : lreduces n1 h1 e1 n2 h2 e2),
  lreduces n h e n2 h2 e2.
Hint Resolve lreduces_step.

Inductive ldiverges_gen diverges n h e : Prop :=
| ldiverges_step : forall n' h' e' 
    (STEP: lreduce n h e n' h' e')
    (DIV: diverges n' h' e': Prop), 
  ldiverges_gen diverges n h e.
Hint Constructors ldiverges_gen.

Definition ldiverges n h e := paco3 ldiverges_gen bot3 n h e.
Hint Unfold ldiverges.
Lemma ldiverges_gen_mon: monotone3 ldiverges_gen. Proof. pmonauto. Qed.
Hint Resolve ldiverges_gen_mon : paco.

(*----------------------------------------------------------------------------*)
(** * Various relations *)
(*----------------------------------------------------------------------------*)

Implicit Types (ty: typ) (v : val) (e: exp) (h : heap).

(** Here are the basic domains that we will be using: *)

Definition vrel := typ -> nat -> val -> nat -> val -> Prop.
Definition krel := typ -> nat -> cont -> nat -> cont -> Prop.
Definition know := typ -> nat -> (val+cont) -> nat -> (val+cont) -> Prop.
Definition hrel := heap -> heap -> Prop.
Definition erel := nat -> exp -> nat -> exp -> Prop.

Notation "'!V' r" := 
  (fun (ty: typ) (n1: nat) (v1: val) (n2: nat) (v2: val) => 
       (r: know) ty n1 (inl v1) n2 (inl v2)) (at level 0).

Notation "'!K' r" := 
  (fun (ty: typ) (n1: nat) (K1: cont) (n2: nat) (K2: cont) => 
       (r: know) ty n1 (inr K1) n2 (inr K2)) (at level 0).

Hint Unfold know vrel erel hrel.

Definition vk2exp (vk: val+cont) : exp :=
  match vk with
  | inl v => v
  | inr k => Vcont k
  end.

Definition know_closed (r : know) := 
  forall ty n vk n' vk' (V: r ty n vk n' vk') x,
  ~ (fv_exp (vk2exp vk) x \/ fv_exp (vk2exp vk') x).
Hint Unfold know_closed.

(*----------------------------------------------------------------------------*)
(** * Value Closure *)
(*----------------------------------------------------------------------------*)

(** The following definition, [vclos], lifts the definition of value 
    equivalence from flexible types only to all types. *)

Inductive vclos_gen (r: know) (vr: vrel) : vrel :=
  | vclos_gen_unit : forall n n', vclos_gen r vr Tunit n Vunit n' Vunit
  | vclos_gen_int  : forall n n' m, vclos_gen r vr Tint n (Vint m) n' (Vint m)
  | vclos_gen_bool_true : forall n n', vclos_gen r vr Tbool n Vtrue n' Vtrue
  | vclos_gen_bool_false : forall n n', vclos_gen r vr Tbool n Vfalse n' Vfalse
  | vclos_gen_prod:
    forall ty1 n1 v1 n1' v1' (VR1: vr ty1 n1 v1 n1' v1') 
           ty2 n2 v2 n2' v2' (VR2: vr ty2 n2 v2 n2' v2')
           n n',
      vclos_gen r vr (Tprod ty1 ty2) n (Vpair v1 v2) n' (Vpair v1' v2')
  | vclos_gen_sum_inl:
    forall ty1 n1 v1 n1' v1' (VR: vr ty1 n1 v1 n1' v1') ty2 n n',
      vclos_gen r vr (Tsum ty1 ty2) n (Vinl v1) n' (Vinl v1')
  | vclos_gen_sum_inr:
    forall ty2 n2 v2 n2' v2' (VR: vr ty2 n2 v2 n2' v2') ty1 n n',
      vclos_gen r vr (Tsum ty1 ty2) n (Vinr v2) n' (Vinr v2')
  | vclos_gen_mu:
    forall ty n1 v n1' v' (VR: vr (tsubst_bvar 0 (Tmu ty) ty) n1 v n1' v') n n',
      vclos_gen r vr (Tmu ty) n (Vroll v) n' (Vroll v')
  | vclos_gen_exists:
    forall ty' ty n1 v n1' v' (VR: vr (tsubst_bvar 0 ty' ty) n1 v n1' v') (WFty': lc_typ 0 ty') n n',
      vclos_gen r vr (Texists ty) n (Vpack v) n' (Vpack v')
  | vclos_gen_name : 
    forall aa n v n' v' (VR: !V r (Tname aa) n v n' v'), vclos_gen r vr (Tname aa) n v n' v'
  | vclos_gen_arrow:
    forall ty1 ty2 n v n' v' (VR: !V r (Tarrow ty1 ty2) n v n' v'), vclos_gen r vr (Tarrow ty1 ty2) n v n' v'
  | vclos_gen_forall:
    forall ty n v n' v' (VR: !V r (Tforall ty) n v n' v'), vclos_gen r vr (Tforall ty) n v n' v'
  | vclos_gen_ref:
    forall ty n1 v n1' v' (VR: !V r (Tref ty) n1 v n1' v') n n', vclos_gen r vr (Tref ty) n v n' v'
  | vclos_gen_cont:
    forall ty n1 K n1' K' (VR: !K r ty n1 K n1' K') n n', vclos_gen r vr (Tcont ty) n (Vcont K) n' (Vcont K').
Hint Constructors vclos_gen.

Inductive vclos (r : know) (ty: typ) n1 (v1: val) n2 (v2: val) : Prop :=
| vclos_intro (REL: vclos_gen r (vclos r) ty n1 v1 n2 v2).
Hint Constructors vclos.

Lemma vclos_gen_mon : forall r0 r ty n1 v1 n2 v2 (REL: vclos_gen r0 r ty n1 v1 n2 v2) r' (LE: r <5= r'),
  vclos_gen r0 r' ty n1 v1 n2 v2.
Proof. intros; dependent destruction ty; inversion REL; eauto. Qed.

Lemma vclos_induction : 
  forall 
    r ty n1 v1 n2 v2 (EL: vclos r ty n1 v1 n2 v2)
    P (PF: forall ty n1 v1 n2 v2 (IH: vclos_gen r P ty n1 v1 n2 v2), P ty n1 v1 n2 v2),
  P ty n1 v1 n2 v2.
Proof.
  fix 7.
  intros; apply PF; destruct EL; destruct REL;
  econstructor; eauto; eapply vclos_induction; eauto.
Qed.

Lemma vclos_induction_strong : 
  forall 
    r ty n1 v1 n2 v2 (EL: vclos r ty n1 v1 n2 v2)
    P (PF: forall ty n1 v1 n2 v2 
             (IH: vclos_gen r (fun ty' n1' v1' n2' v2' => P ty' n1' v1' n2' v2' /\ vclos r ty' n1' v1' n2' v2') ty n1 v1 n2 v2), 
           P ty n1 v1 n2 v2),
  P ty n1 v1 n2 v2.
Proof.
  intros; cut (P ty n1 v1 n2 v2 /\ vclos r ty n1 v1 n2 v2); [intuition|].
  revert P PF; eapply (vclos_induction EL); clear ty v1 v2 EL; intros.
  split; eauto using vclos_gen_mon.
  econstructor; eapply vclos_gen_mon; eauto.
  by simpl; intros; hdes; eauto.
Qed.

Lemma vclos_mon: 
  forall r r' ty n v n' v' (VC: vclos r ty n v n' v') (LT: r <5= r'), 
    vclos r' ty n v n' v'.
Proof. 
  intros until v'; intros REL; eapply (vclos_induction REL); intros.
  inv IH; clear IH; (econs; econs; eauto).
Qed.

Lemma vclos_mon_lv: forall g ty n1 v1 n2 v2
    (REL: vclos g ty n1 v1 n2 v2)
    (MONG: forall ty n1 n1' n2 n2' 
             (REL: !V g ty n1 v1 n2 v2) (LEN1: n1 <= n1') (LEN2: n2 <= n2'), 
           !V g ty n1' v1 n2' v2)
    n1' (LEN1: n1 <= n1') n2' (LEN2: n2 <= n2'),
  vclos g ty n1' v1 n2' v2.
Proof. induction 1; destruct REL; i; eauto. Qed.

Lemma vclos_gen_closed : 
  forall r vr ty n v n' v' (VREL : vclos_gen r vr ty n v n' v') 
    (C : know_closed r) 
    (CV: forall ty n v n' v' (V: vr ty n v n' v') x, 
         ~(fv_exp v x \/ fv_exp v' x)) 
  x, ~(fv_exp v x \/ fv_exp v' x).
Proof.
  i; destruct VREL; simpl; subst; try specialize (C _ _ _ _ _ VR); des; simpls;
    try assert (X:= CV _ _ _ _ _ VR); 
    try assert (X:= CV _ _ _ _ _ VR1); 
    try assert (Y:= CV _ _ _ _ _ VR2); 
    esplits; i; des; eauto.
Qed.

Lemma vclos_closed : 
  forall r ty n v n' v' (VREL : vclos r ty n v n' v') (C : know_closed r) x, 
  ~ (fv_exp v x \/ fv_exp v' x).
Proof.
  intros until v'; intro VREL; pattern ty, n, v, n', v'. 
  eapply (vclos_induction VREL); i; eapply vclos_gen_closed in IH; eauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * Dependent World *)
(*----------------------------------------------------------------------------*)

Structure world_dep (pst : Type) (pext : relation pst) := 
  { wnam : typename -> Prop 
  ; wst  : Type
  ; wextpub : relation wst
  ; wval  : (pst -> wst -> know) -> pst -> wst -> know
  ; wheap : (pst -> wst -> know) -> pst -> wst -> hrel
  ; wst_countable   : countable_set wst 
  ; wextpub_refl    : forall x, wextpub x x
  ; wextpub_trans   : forall x y (EXT: wextpub x y) z (EXT': wextpub y z), wextpub x z
  ; wnam_supports   : forall sp s r a n v n' v' 
                        (WV: !V (wval sp s r) (Tname a) n v n' v'), 
                      wnam a 
  ; wval_mon        : forall sp s r ty n vk n' vk' (WV: wval r sp s ty n vk n' vk')
                        r'  (LT: r <7= r')
                        sp' (RP: pext sp' sp) 
                        s'  (R: wextpub s' s)
                        n1  (LN: n <= n1)
                        n1' (LN': n' <= n1'),
                      wval r' sp' s' ty n1 vk n1' vk'
  ; wval_closed     : forall sp s r, know_closed (wval r sp s)
  ; wheap_mon       : forall sp s r h h' (PR: wheap r sp s h h')
                        r' (LT: forall sp s, (* !V *) (r sp s) <5= (* !V *) (r' sp s)),
                      wheap r' sp s h h'
  }.
Hint Immediate wextpub_refl.

Structure world_data pst := mkwd
{ dwst  : Type
; dwnam : typename -> Prop
; dwextpub : relation dwst
; dwheap : (pst -> dwst -> know) -> pst -> dwst -> hrel }.

Coercion w2d pst (pext: relation pst) (w: world_dep pext) : world_data pst :=
 {| dwst := wst w; dwnam := wnam w;
    dwextpub := wextpub w; dwheap := wheap w |}.

Lemma dwextpub_unfold: forall pst pext (w: @world_dep pst pext) s s',
  wextpub w s s' -> dwextpub w s s'.
Proof. auto. Qed.
Hint Resolve dwextpub_unfold.

(*----------------------------------------------------------------------------*)
(** * Full World *)
(*----------------------------------------------------------------------------*)

Definition world := world_dep (fun x y: unit => True).

(*----------------------------------------------------------------------------*)
(** * Ref World *)
(*----------------------------------------------------------------------------*)

(** We define a ref world that defines the meaning of the
    mutable reference type. *)

Definition wrefst (s : typ -> location -> location -> Prop) :=
     (forall ty loc1 loc2 (IN : s ty loc1 loc2), <<LC: lc_typ 0 ty >>)
  /\ (forall ty loc1 loc2 (IN : s ty loc1 loc2) ty' loc1' loc2' (IN' : s ty' loc1' loc2'),
        (forall (EQ: loc1 = loc1'), <<TEQ: ty = ty' >> /\ <<LEQ: loc2 = loc2' >>) /\
        (forall (EQ: loc2 = loc2'), <<TEQ: ty = ty' >> /\ <<LEQ: loc1 = loc1' >>))
  /\ (exists tydom ldom, 
       forall ty l1 l2 (IN: s ty l1 l2), 
         <<INTY: In ty tydom>> /\ <<INL1: In l1 ldom>> /\ <<INL2: In l2 ldom>>).

Fixpoint list_prod A B (la: list A) (lb: list B) : list (A*B) :=
  match la with
  | nil => nil
  | a :: la' => map (fun b => (a,b)) lb ++ list_prod la' lb
  end.

Definition wrefst_to_list (s: {s | wrefst s}) : list (typ*location*location) :=
  let (tydom, P) := cid (proj2 (proj2(proj2_sig s))) in
  let (ldom, _)  := cid P in
  filter 
   (fun x => if LEM(`s (fst(fst x)) (snd(fst x)) (snd x)) then true else false)
   (list_prod (list_prod tydom ldom) ldom).

Lemma list_prod_in: forall A B (la: list A) (lb: list B) a b
    (INA: In a la) (INB: In b lb),
  In (a,b) (list_prod la lb).
Proof.
  induction la; [done|].
  s; i; des; subst; apply in_or_app; [|eauto].
  left; induction lb; [done|].
  ss; des; subst; eauto.
Qed.

Lemma wrefst_countable: countable_set (sig wrefst).
Proof.
  destruct (countable_list (countable_prod
            (countable_prod typ_countable nat_countable) nat_countable)) as (f, INJ).
  exists (fun s => f (wrefst_to_list s)).
  intros (s, PS) (s',PS') EQ; apply INJ in EQ; clear f INJ.
  apply sig_eq; extensionalities ty l1 l2.
  revert EQ; unfold wrefst_to_list.
  repeat gen_cid; intros (tydom' & X) (tydom & Y).
  repeat gen_cid; intros (ldom' & DOM') (ldom & DOM) EQ; clear X Y.
  ss; apply prop_ext; split; intro SIN.
  - match goal with [EQ: ?t = _ |-_] => assert (IN: In (ty,l1,l2) t) end.
      apply filter_In; s; split.
        by assert (X:=DOM _ _ _ SIN); des; eauto using list_prod_in.
      by destruct (LEM _).
    rewrite EQ in IN; apply filter_In in IN; des.
    by ss; destruct (LEM _).
  - match goal with [EQ: _ = ?t |-_] => assert (IN: In (ty,l1,l2) t) end.
      apply filter_In; s; split.
        by assert (X:=DOM' _ _ _ SIN); des; eauto using list_prod_in.
      by destruct (LEM _).
    rewrite <-EQ in IN; apply filter_In in IN; des.
    by ss; destruct (LEM _).
Qed.

Definition wref_val s ty n (vk: val+cont) n' (vk': val+cont)
  := exists loc loc' ty',
     <<NLT: n > 0>> /\ <<NLT': n' > 0>>
  /\ <<TEQ: ty = Tref ty'>> 
  /\ <<VEQ: vk = inl (Vloc loc) >> 
  /\ <<VEQ': vk' = inl (Vloc loc')>> 
  /\ <<TLL: s ty' loc loc'>>.

Definition wref_heap r (s: sig wrefst) h h' := 
  <<HSAT : forall ty loc loc' (R: `s ty loc loc'), 
           exists n v n' v', 
           <<HL: h loc = Some v>> /\ <<HL': h' loc' = Some v'>> /\
           <<VC: vclos (r () s) ty n v n' v'>> >> /\
  <<HDOM1 : forall l v (IN: h l = Some v), exists ty l', `s ty l l' >> /\
  <<HDOM2 : forall l' v' (IN: h' l' = Some v'), exists ty l, `s ty l l' >>.

Program Definition wref : world :=
  {| wnam n := False
   ; wst := { s : typ -> location -> location -> Prop | wrefst s } 
   (* ; wext s s' := s' <3= s *)
   ; wextpub s s' := s' <3= s
   ; wval r u s ty n vk n' vk' := wref_val s ty n vk n' vk'
   ; wheap r u s h h' := wref_heap r s h h'
  |}.
Next Obligation. apply wrefst_countable. Qed.
Next Obligation. by unfold wref_val in *; des. Qed.
Next Obligation. unfold wref_val in *; des; unnw; esplits; eauto; omega. Qed.
Next Obligation. unfold wref_val in *; des; subst; eauto. Qed.
Next Obligation.
rdes PR; unnw; r; esplits; ii; eauto. 
eapply HSAT in R; des; eauto 10 using vclos_mon. 
Qed.

Program Definition sref0 : wst wref := exist _ (fun _ _ _ => False) _.
Next Obligation.
  red; repeat (split; intros; vauto).
  Existential 1 := nil. Existential 1 := nil.
Qed.

(*----------------------------------------------------------------------------*)
(** * Local World *)
(*----------------------------------------------------------------------------*)

Definition world_local := world_dep (wextpub wref).

Inductive gmon pst pext (w: @world_dep pst pext) (g: wst w -> know) s : know :=
| gmon_vk: forall s' ty n1 n1' vk1 n2 n2' vk2
    (REL: g s' ty n1' vk1 n2' vk2)
    (EXT: wextpub w s s') (LN1: n1' <= n1) (LN2: n2' <= n2),
  gmon w g s ty n1 vk1 n2 vk2
.
Hint Constructors gmon.

Lemma gmon_mon: forall pst pext w g s ty n1 vk1 n2 vk2
    (REL: @gmon pst pext w g s ty n1 vk1 n2 vk2)
    g' (LE: g <6= g')
    s' (EXT: wextpub w s' s)
    n1' (LN1: n1 <= n1') n2' (LN2: n2 <= n2'),
  gmon w g' s' ty n1' vk1 n2' vk2.
Proof.
  i; destruct REL; econs; eauto using wextpub_trans; try omega.
Qed.

(*----------------------------------------------------------------------------*)
(** * Product of two local worlds *)
(*----------------------------------------------------------------------------*)

Program Definition wprod (w1 w2 : world_local) : world_local :=
  {| wnam := nunion (wnam w1) (wnam w2)
   ; wextpub s s' := <<WEP: wextpub w1 (fst s) (fst s')>> /\ <<WEP': wextpub w2 (snd s) (snd s')>>
   ; wval r sp s ty n vk n' vk' := 
     wval w1 (fun sp s1 => gmon w2 (fun s2 => r sp (s1,s2)) (snd s)) sp (fst s) ty n vk n' vk' \/ 
     wval w2 (fun sp s2 => gmon w1 (fun s1 => r sp (s1,s2)) (fst s)) sp (snd s) ty n vk n' vk'
   ; wheap r sp s h h' := exists h1 h1' h2 h2', 
       <<WH:  wheap w1 (fun sp s1 => gmon w2 (fun s2 => r sp (s1,s2)) (snd s)) sp (fst s) h1 h1'>> /\
       <<WH': wheap w2 (fun sp s2 => gmon w1 (fun s1 => r sp (s1,s2)) (fst s)) sp (snd s) h2 h2'>> /\
       <<HP: hplus h1 h2 h>> /\ <<HP': hplus h1' h2' h'>> |}.
Next Obligation. eapply countable_prod; eauto using wst_countable. Qed.
Next Obligation. simpls; des; eauto 10 using (* wext_trans, *) wextpub_trans. Qed. 
Next Obligation. by des; constructor (eby eapply wnam_supports). Qed.
Next Obligation. 
  des; [left|right]; eapply wval_mon; eauto; inst; s; eauto using gmon_mon.
Qed.
Next Obligation. 
  des; eapply wval_closed; eauto; s; ii; destruct V0; subst; eapply C; eauto.
Qed.
Next Obligation. 
  des; esplits; eauto; eapply wheap_mon; eauto; inst; s; eauto using gmon_mon.
Qed.

(*----------------------------------------------------------------------------*)
(** * Lifting of a local world *)
(*----------------------------------------------------------------------------*)

Definition is_ref (vk vk': val+cont) (ty: typ) : bool :=
  match vk, vk', ty with
  | inl _, inl _, Tref _ => true
  | _, _, _ => false
  end.

Definition wlift_val (w: world_local) r s ty n vk n' vk' :=
  if is_ref vk vk' ty 
  then wval wref (fun _ _ _ _ _ _ _ => False) () (fst s) ty n vk n' vk'
  else wval w (fun sr s => r () (sr, s)) (fst s) (snd s) ty n vk n' vk'.

Definition wlift_heap (w: world_local) r s h h' :=
  exists h1 h1' h2 h2', 
  <<WH: wheap wref (fun _ sr => r () (sr, snd s)) () (fst s) h1 h1'>> /\
  <<WH': wheap w (fun sr s => r () (sr,s)) (fst s) (snd s) h2 h2'>> /\
  <<HP: hplus h1 h2 h>> /\ <<HP': hplus h1' h2' h'>>.

Program Definition wlift (w : world_local) : world :=
{| wnam := nunion (wnam wref) (wnam w)
 ; wextpub s s' := <<WREP: wextpub wref (fst s) (fst s')>> /\ <<WREP': wextpub w (snd s) (snd s')>>
 ; wval r sp s ty n vk n' vk' := wlift_val w r s ty n vk n' vk'
 ; wheap r sp s h h' := wlift_heap w r s h h'
|}.
Next Obligation. eapply countable_prod; eauto using wst_countable, (wst_countable wref). Qed.
Next Obligation. des; eauto 10 using wextpub_trans. Qed.
Next Obligation. rdes WV; ss; constructor (eby eapply wnam_supports). Qed.
Next Obligation. unfold wlift_val in *; des; destruct (is_ref _ _ _); eauto using wval_mon. Qed.
Next Obligation. unfold wlift_val in *; destruct (is_ref _ _ _); eapply wval_closed; eauto. Qed.
Next Obligation. 
unfold wlift_heap in *; ss; unfold wref_heap in *.
des; ss; do 4 eexists; repeat split; try eapply wheap_mon; des; unnw; eauto.
by intros; specialize (HSAT _ _ _ R); des; eauto 10 using vclos_mon.
Qed.

(*----------------------------------------------------------------------------*)
(** * Global Knowledge *)
(*----------------------------------------------------------------------------*)

(** A custom order between two value relations that requires the
    meaning of reference and abstract types to agree between the
    two relations. *)

Structure lerefn N (r r' : know) : Prop :=
{ refn_le : forall ty n vk n' vk' (R: r ty n vk n' vk'), r' ty n vk n' vk'
; refn_ref: forall ty n v n' v' (R: r' (Tref ty) n (inl v) n' (inl v')),
            r (Tref ty) n (inl v) n' (inl v')
; refn_nam: forall a n v n' v' (NM: N a) (R: r' (Tname a) n (inl v) n' (inl v')), 
            r (Tname a) n (inl v) n' (inl v') 
}.
Hint Constructors lerefn.

(** A global value relation [g] is consistent w.r.t. a world [w] if
    it is (i) well-formed; 
        (ii) monotonic w.r.t state transitions 
    and (iii) relates any two values that are related by [w]. *)

Structure wglobgen (w: world_data ()) 
  (lk: (dwst w -> know) -> dwst w -> know) (g : dwst w -> know) : Prop :=
{ wg_clo : forall s, know_closed (g s)
; wg_mon : forall s ty n vk n' vk' (WV: g s ty n vk n' vk')
             n1  (LN: n <= n1)
             n1' (LN': n' <= n1')
             s' (R: dwextpub w s' s),
           g s' ty n1 vk n1' vk'
; wg_le: forall s, lerefn (dwnam w) (lk g s) (g s)
; wg_sc: forall s n1 n2, !K (g s) Tint n1 [] n2 []
}.
Hint Constructors wglobgen.

Definition wglob (w: world) g := 
  wglobgen w (fun r => wval w (fun _=>r) ()) g.
Arguments wglob /.

(*----------------------------------------------------------------------------*)
(** * Expression and Continuation Equivalence *)
(*----------------------------------------------------------------------------*)

(** 
   Co-inductive definition of when two (closed) expressions are related. 
*)

Inductive wstuck (Rf : know) (Rv : know) n1 e1 n2 e2 : Prop :=
  | wstuck_Return
      ty nk1 nk2 K1 K2 nv1 nv2 (v1 v2: voe)
      (EQ1  : e1 = app_cont K1 (e_of v1))
      (EQ2  : e2 = app_cont K2 (e_of v2))
      (CREL : Rf ty nk1 (inr K1) nk2 (inr K2))
      (VREL : vclos Rv ty nv1 (v_of v1) nv2 (v_of v2))
      (NEQ1 : n1 >= nv1 + nk1)
      (NEQ2 : n2 >= nv2 + nk2)
  | wstuck_App
      ty ty' nv1 nv2 (f1 v1 f2 v2 : voe) nk1 K1 nk2 K2
      (EQ1  : e1 = app_cont K1 (Eapp (e_of f1) (e_of v1)))
      (EQ2  : e2 = app_cont K2 (Eapp (e_of f2) (e_of v2)))
      (CREL : Rf (Tarrow ty ty') n1 (inl (v_of f1)) n2 (inl (v_of f2)))
      (VREL : vclos Rv ty nv1 (v_of v1) nv2 (v_of v2))
      (KREL : Rv ty' nk1 (inr K1) nk2 (inr K2))
  | wstuck_Inst
      ty ty' (f1 f2 : voe) nk1 K1 nk2 K2
      (EQ1  : e1 = app_cont K1 (Einst (e_of f1)))
      (EQ2  : e2 = app_cont K2 (Einst (e_of f2)))
      (LC   : lc_typ 0 ty')
      (CREL : Rf (Tforall ty) n1 (inl (v_of f1)) n2 (inl (v_of f2)))
      (KREL : Rv (tsubst_bvar 0 ty' ty) nk1 (inr K1) nk2 (inr K2))
.
Hint Constructors wstuck.

Inductive wexp_core (w : world_data ()) (r g: dwst w -> know) (s: dwst w) 
  (hf hh: heap) (n: nat) (e:exp) (hf' hh': heap) (n': nat) (e': exp) : Prop :=
| wexp_Diverge
    (DIV : ldiverges n hh e) 
    (DIV': ldiverges n' hh' e')
| wexp_Stuck
     s' (EXT: dwextpub w s' s)
     e1 e2 ne1 ne2 (EREL: wstuck (r s') (g s') ne1 e1 ne2 e2)
     h1 h2 (HREL: dwheap w (fun _ => g) () s' h1 h2)
     hh1 (PLUS1: hplus h1 hf hh1)
     hh2 (PLUS2: hplus h2 hf' hh2)
     (RED: lreduces n hh e ne1 hh1 e1)
     (RED': lreduces n' hh' e' ne2 hh2 e2)
.
Hint Constructors wexp_core.

Definition wexpgen (w: world_data ()) (r g: dwst w -> know) (s: dwst w) (n:nat) e (n':nat) e' :=
  forall 
    h h' (HREL: dwheap w (fun _ => g) () s h h')
    hf hh (HPLUS: hplus h hf hh)
    hf' hh' (HPLUS': hplus h' hf' hh'),
    wexp_core w r g s hf hh n e hf' hh' n' e' : Prop.

Definition wexp w g := wexpgen w g g.

(*----------------------------------------------------------------------------*)
(** * Program Equivalence *)
(*----------------------------------------------------------------------------*)

Definition wstable (w : world_local) :=
  forall 
    g (gOK: wglob (wlift w) g)
    sr s h1 h2 (HR: wheap w (fun sr s => g (sr,s)) sr s h1 h2)
    sr' (EXT: wextpub _ sr' sr)
    hs1 hs2 (SHR: wheap wref (fun _ sr => g (sr,s)) () sr' hs1 hs2)
    hh1 (DIS: hplus hs1 h1 hh1)
    hh2 (DIS: hplus hs2 h2 hh2),
  exists s', <<WEP: wextpub _ s' s>> /\ <<WH: wheap _ (fun sr s => g(sr,s)) sr' s' h1 h2>>.

(** We say that a world is inhabited iff it relates the empty heap to itself. *)

Definition winhabitedgen (w : world_data ()) lk := 
  exists s0, forall g (gOK: wglobgen w lk g), 
  dwheap w (fun _ => g) () s0 hemp hemp.

Definition winhabited (w: world) :=
  winhabitedgen w (fun r => wval w (fun _=>r) ()).

(** A world is consistent if it any functions it declares as 
    being equivalent are indeed so when applied to equivalent
    arguments. *)

Inductive drop (r : erel) n e n' e' : Prop :=
_drop n1 e1 n1' e1'
  (R: r n1 e1 n1' e1')
  (LRED: forall h, lreduce n h e n1 h e1)
  (LRED': forall h, lreduce n' h e' n1' h e1')
.

Definition wconsistentgen (w : world_data ()) lk :=
  forall g s (gOK: wglobgen w lk g),
  wstuck (lk g s) (g s) <4= drop (wexp w g s).

Definition wconsistent (w : world) :=
  wconsistentgen w (fun r => wval w (fun _=>r) ()).

(** Two expressions are related by a world [w] in a typing environment [D; G],
    if they are related by [w] after substituting any pairwise related values
    for the expressions' free variables. *)

Structure wfdg G r dd gg : Prop :=
{ wfdg_ddok: forall a, lc_typ 0 (dd a)
; wfdg_ggok: forall (x: termvar),
             vclos r (tsubst_name dd (G x))
             (((fst<*>fst)<*>gg) x) (((snd<*>fst)<*>gg) x) (((fst<*>snd)<*>gg) x) (((snd<*>snd)<*>gg) x)
}.
Hint Constructors wfdg.
Create HintDb wfdg.
Hint Resolve wfdg_ddok wfdg_ggok : wfdg.

Inductive wequivgen (w: world_data ()) lk G ty e1 e2 : Prop :=
  wequiv_intro nf1 nf2
    (WEXP: forall s
       g (gOK: wglobgen w lk g)
       dd gg (dgOK: wfdg G (g s) dd gg)
       nk1 K1 nk2 K2 (KOK: g s (tsubst_name dd ty) nk1 (inr K1) nk2 (inr K2)),
     wexp w g s
     (nf1 ((fst<*>fst)<*>gg) + nk1)
     (app_cont K1 (subst_exp ((exp_of_val<*>snd<*>fst)<*>gg) e1))
     (nf2 ((fst<*>snd)<*>gg) + nk2)
     (app_cont K2 (subst_exp ((exp_of_val<*>snd<*>snd)<*>gg) e2))).
Hint Constructors wequivgen.

Definition wequiv (w: world) := 
  wequivgen w (fun r => wval w (fun _=>r) ()).
Hint Unfold wequiv.

Definition weq w G ty e1 e2 : Prop :=
  <<INH : winhabited w>> /\
  <<CS : wconsistent w>> /\
  <<EQ : wequiv w G ty e1 e2>>.
Hint Unfold weq.

(** Finally, two expressions are equivalent at a given type, if there 
    exists an inhabited, non-ref world, whose product with the global
    [wref] world is consistent and relates the two expressions. *)

Structure wfprog G ty e1 e2 : Prop :=
{ wfprog_envok: forall x: termvar, lc_typ 0 (G x)
; wfprog_tyok: lc_typ 0 ty 
; wfprog_e1: wf_exp e1
; wfprog_e2: wf_exp e2
; wfprog_finite: exists xs, forall x (NOTIN: ~ In x xs), G x = Tunit
}.
Hint Constructors wfprog.
Create HintDb wfprog.
Hint Resolve wfprog_envok wfprog_tyok wfprog_e1 wfprog_e2 : wfprog.

Definition wreq G ty e1 e2 :=
  <<WFP: wfprog G ty e1 e2>> /\
  <<REQ: forall N (INF: countably_infinite N), 
         exists w : world_local,
         <<NAM: wnam w <1= N >> /\
         <<WSTA: wstable w >> /\
         <<WEQ: weq (wlift w) G ty e1 e2>> >>.
