Require Import Program vgtac hpattern.
Require Import FunctionalExtensionality Compare_dec Le Image List Ensembles Mult.
Require Import Classical IndefiniteDescription ChoiceFacts.
Require Import lang.

Set Implicit Arguments.
Unset Strict Implicit.
Open Scope string_scope.

(*----------------------------------------------------------------------------*)
(** ** Classical Logic *)
(*----------------------------------------------------------------------------*)

(** We assume classical logic, functional and propositional extensionality. *)

Lemma prop_ext : ClassicalFacts.prop_extensionality.
Proof.
red; intros; exploit (@Extensionality_Ensembles () (fun _ => A) (fun _ => B)).
  by split; red; intros; unfold In in *; intuition.
by intros EQ; change A with ((fun _ => A) ()); change B with ((fun _ => B) ()); rewrite EQ.
Qed.

Lemma proof_irr: ClassicalFacts.proof_irrelevance.
Proof. exact (ClassicalFacts.ext_prop_dep_proof_irrel_cic prop_ext). Qed.

Lemma proj1_sig_inj: forall A (N : A -> Prop) (a a': sig N) (EQ: `a = `a'), a = a'.
Proof.
intros; destruct a, a'; simpls; subst; f_equal; eauto using proof_irr.
Qed.

Lemma sig_eta: forall A (P: A -> Prop) (x : sig P) pf, exist P (`x) pf = x.
Proof. by intros; destruct x; simpls; apply proj1_sig_inj. Qed.

Lemma LEM: forall (P: Prop), {P} + {~P}.
Proof.
intros; eapply constructive_definite_descr_excluded_middle; eauto using classic.
repeat intro; eapply constructive_indefinite_description in H.
by destruct H as (a & H); destruct H; eauto.
Qed.

Lemma or2sum: forall P Q (OR: P \/ Q), {P} + {Q}.
Proof.
intros; destruct (LEM P); [by left|]; destruct (LEM Q); [by right|].
by exfalso; intuition.
Qed.

Ltac gen_cid :=
  match goal with [|- context[@constructive_indefinite_description ?A ?B ?C]] =>
    generalize (@constructive_indefinite_description A B C)
  end.

Ltac gen_eqrefl :=
  match goal with [|-context[@eq_refl ?A ?B]] => 
    generalize (@eq_refl A B) 
  end.

(*----------------------------------------------------------------------------*)
(** ** Countably infinite sets *)
(*----------------------------------------------------------------------------*)

Definition surjective A B (f: A -> B) : Prop := forall b, exists a, b = f a.

Definition countably_infinite_set (A: Type) :=
  exists f: A -> nat, exists g: nat -> A,
    (forall a, g (f a) = a) /\
    (forall n, f (g n) = n).

Definition countably_infinite_set2 (A:Type) :=
  exists f: A -> nat, exists g: nat -> A,
    injective _ _ f /\ injective _ _ g.

Definition countably_infinite A (N: A -> Prop) :=
  countably_infinite_set (sig N).

Definition ndisj A (N1 N2: A -> Prop) := forall n, N1 n -> N2 n -> False.

Definition nunion A (N1 N2: A -> Prop) := fun n => N1 n \/ N2 n. 

Lemma ndisj_mon : 
  forall 
    A N1' N2' (DIS: ndisj (A:=A) N1' N2')
    (N1: A->Prop) (LT1: forall a (EL: N1 a), N1' a)
    (N2: A->Prop) (LT2: forall a (EL: N2 a), N2' a),
  ndisj N1 N2.
Proof. red; intros; eauto. Qed.

Lemma even_or_odd: forall n, exists i, n = 2 * i \/ n = S (2 * i).
Proof. by intros; induction n; des; subst; eauto; exists (S i); omega. Qed.

(** nat_pair *)

Fixpoint exponential (n m: nat) : nat :=
  match m with | 0 => 1 | S m' => exponential n m' * n end.

Fixpoint nat_pair (nm: nat*nat) : nat := exponential 2 (fst nm) * exponential 3 (snd nm).

Lemma exponential2_inv: forall n m, exponential 2 n = m * 3 -> False.
Proof.
  induction n; intros; simpls; [by omega|].
  by destruct (even_or_odd m); des; subst; [specialize (IHn x)|]; omega.
Qed.

Lemma exponential2_inj: injective _ _ (exponential 2).
Proof.
  intro n; induction n; intros []; intros; simpls; try omega.
  by f_equal; eapply IHn; omega.
Qed.

Lemma nat_pair_inj: injective _ _ nat_pair.
Proof.
  intros (n & m); revert n; induction m; intros; destruct y; simpls.
    destruct n1; simpls.
      by rewrite !mult_1_r in H; eapply exponential2_inj in H; rewrite H.
    by exfalso; rewrite mult_1_r, mult_assoc in H; eauto using exponential2_inv.
  destruct n1; simpls.
    by exfalso; rewrite mult_1_r, mult_assoc in H; eauto using exponential2_inv.
  exploit (IHm n (n0,n1)).
    simpls; rewrite !mult_assoc in H; omega.
  by intros X; inv X.
Qed.

(** Equivalence of countably_infinite_set and countably_infinite_set2 *)

Lemma inj_srj_impl_bij: forall A B (f: A -> B)
    (INJ: injective _ _ f)
    (SRJ: surjective f),
  exists g, (forall a, g (f a) = a) /\ (forall b, f (g b) = b).
Proof. by intros; eapply functional_choice in SRJ; des; eauto. Qed.

Fixpoint count_prop (P: nat -> Prop) (n: nat) : nat :=
  match n with
  | 0 => 0
  | S n' => match LEM (P n') with | left _ => 1 | right _ => 0 end + count_prop P n'
  end.

Definition compactify A (f: A -> nat) (a : A) : nat :=
  count_prop (fun n => exists a, n = f a) (f a).

Lemma count_prop_mon: forall P n n' (LE: n <= n'), count_prop P n <= count_prop P n'. 
Proof. by intros; induction LE; simpl; omega. Qed.

Lemma compactify_le: 
  forall A (f: A -> nat) a n (LE: n <= compactify f a), exists a', n = compactify f a'.
Proof.
unfold compactify; intros; eapply le_lt_eq_dec in LE; destruct LE as [LE|LE]; eauto.
revert LE; generalize (f a); clear a; intros m LE.
induction m; intros; simpls; [by exfalso; omega|].
destruct LEM; eauto; []; eapply le_lt_eq_dec in LE; destruct LE.
- by eapply IHm; omega.
- by des; subst; exists a; omega.
Qed.

Lemma nat_map_inj_impl_unbounded: 
  forall (f: nat -> nat) (INJ: injective _ _ f) m, exists n, f n >= m.
Proof.
  intros; eapply NNPP; intro X; assert (Y := not_ex_all_not _ _ X); clear X; simpls.
  assert (CARDSEG: forall m, cardinal _ (fun n => n < m) m).
  {
    clear; induction m.
      setoid_rewrite (@Extensionality_Ensembles nat); [by econstructor|].
      by split; repeat intro; unfold In in *; [exfalso; omega|destruct H].
    setoid_rewrite (@Extensionality_Ensembles nat).
    - by econstructor; eauto; unfold In; instantiate (1:= m); omega.
    - split; repeat intro; unfold In in *.
        assert (X: x = m \/ x < m) by omega.
        by des; subst; [econstructor 2| econstructor 1].
      by destruct H; red in H; [omega|inv H].
  }
  assert (CARDIM := @injective_preserves_cardinal nat nat (fun n => n < S m) _ _ INJ (CARDSEG (S m))).
  assert (INCL: Included nat (Im _ _ (fun n => n < S m) f) (fun n => n < m)).
    by repeat intro; inv H; unfold In in *; specialize (Y x0); omega.
  assert (CARD: exists n, cardinal nat (Im nat nat (fun n => n < S m) f) n).
    by eauto using finite_cardinal, cardinal_finite, Finite_downward_closed.
  des; specialize (CARDIM _ CARD); subst.
  assert (S m <= m) by eauto using incl_card_le.
  by omega.
Qed.

Lemma compactify_inj: 
  forall A (f: A -> nat) (INJ: injective _ _ f), injective _ _ (compactify f).
Proof.
red; intros; eapply INJ, NNPP; red; intros; unfold compactify in *.
destruct (lt_eq_lt_dec (f x) (f y)) as [[CMP|CMP]|CMP]; try contradiction.
- assert (LE: S (f x) <= f y) by omega.
  eapply (count_prop_mon (fun n => exists a, n = f a)) in LE.
  by rewrite <-H in LE; simpl in LE; destruct LEM; [omega|eauto].
- assert (LE: S (f y) <= f x) by omega.
  eapply (count_prop_mon (fun n => exists a, n = f a)) in LE.
  by rewrite H in LE; simpl in LE; destruct LEM; [omega|eauto].  
Qed.

Lemma countably_infinite_set2_impl_set: 
  forall A (CNT: countably_infinite_set2 A),
  countably_infinite_set A.
Proof.
intros; rdes CNT; exists (compactify f); eapply inj_srj_impl_bij.

Case "Injectivity".
eauto using compactify_inj.

Case "Surjectivity".
red; intros; unfold compactify in *; induction b.
- by eapply compactify_le with (a:= g 0); omega.
- des; edestruct (@nat_map_inj_impl_unbounded (fun n => compactify f (g(n)))).
    by red; intros; eapply compactify_inj in H; eauto.
  by subst; eapply compactify_le; eauto.
Qed.

Lemma countably_infinite_set_impl_set2: 
  forall A (CNT: countably_infinite_set A),
  countably_infinite_set2 A.
Proof.
intros; rdes CNT; exists f g; split; red; intros.
- by rewrite <-(CNT x), <-(CNT y), H.
- by rewrite <-(CNT0 x), <-(CNT0 y), H.
Qed.

(** A countably infinite set can be divided into two disjoint countably 
    infinite sets. *)

Lemma countably_infinite_split: 
  forall A (N: A -> Prop) (INF: countably_infinite N),
    exists N1 N2, 
      <<Nci1: countably_infinite N1>> /\ <<Nci2: countably_infinite N2>> /\
      <<Nunion: N = nunion N1 N2>> /\ <<Ndisj: ndisj N1 N2>>.
Proof.
  intros; do 2 red in INF; des.
  set (N1 a := exists i, a = proj1_sig (g (2 * i))).
  set (N2 a := exists i, a = proj1_sig (g (S(2 * i)))).
  set (f1 (a : sig N1) :=
    projT1 (constructive_indefinite_description _ (proj2_sig a))).
  set (f2 (a : sig N2) :=
    projT1 (constructive_indefinite_description _ (proj2_sig a))).
  set (g1 i := exist N1 (proj1_sig (g (2 * i))) (ex_intro _ _ eq_refl)).
  set (g2 i := exist N2 (proj1_sig (g (S(2 * i)))) (ex_intro _ _ eq_refl)).

  assert (F1F: forall a Pa1 Pa, f (exist _ a Pa) = 2 * (f1 (exist _ a Pa1))).
    intros; unfold f1; gen_cid; intros (n & Pn); simpls.
    by rewrite <- INF0; f_equal; subst; eapply proj1_sig_inj.
  assert (F2F: forall a Pa1 Pa, f (exist _ a Pa) = S(2 * (f2 (exist _ a Pa1)))).
    intros; unfold f2; gen_cid; intros (n & Pn); simpls.
    by rewrite <- INF0; f_equal; subst; eapply proj1_sig_inj.

  exists N1 N2; split.
    exists f1 g1; split; intros. 
      destruct a as (a & Pa); unfold g1. 
      assert (Pa' : N a) by (rdes Pa; subst; apply (proj2_sig (g(2*i)))).
      gen_eqrefl; hrewrite <-(F1F _ Pa Pa'); rewrite INF.
      by simpls; eauto using proj1_sig_inj.
    unfold f1, g1; gen_cid; intros (n' & Pn'); simpls.
    apply proj1_sig_inj in Pn'.
    cut (n+(n+0)=n'+(n'+0)); [by omega|].
    by rewrite <-INF0; rewrite <-INF0 at 1; rewrite Pn'.
  split.
    exists f2 g2; split; intros.
      destruct a as (a & Pa); unfold g2. 
      assert (Pa' : N a) by (rdes Pa; subst; apply (proj2_sig (g(S(2*i))))).
      gen_eqrefl; hrewrite <-(F2F _ Pa Pa'); rewrite INF.
      by simpls; eauto using proj1_sig_inj.      
    unfold f2, g2; gen_cid; intros (n' & Pn'); simpls.
    apply proj1_sig_inj in Pn'.
    cut (S(n+(n+0))=S(n'+(n'+0))); [by omega|].
    by rewrite <-INF0; rewrite <-INF0 at 1; rewrite Pn'.
  split.
  - red; extensionality a; apply prop_ext; unfold nunion, N1, N2; split; intros X.
    + assert (Y := X); change a with (proj1_sig(exist _ _ X)) in Y.
      rewrite <- (INF (exist _ _ X)) in Y.
      destruct (even_or_odd (f (exist _ _ X))); des; [left|right];
       by exists x; rewrite <- H; rewrite INF.
    + by des; subst; apply proj2_sig.
  - intros n (i & X) (j & Y); subst. 
    eapply proj1_sig_inj in Y.
    cut (2*i = S(2*j)); [by omega|].
    by rewrite <- INF0; rewrite <- INF0 at 1; rewrite Y.
Qed.

Lemma countably_infinite_plus_elmt: 
  forall A (N: A -> Prop) (INF: countably_infinite N) (el: A),
    countably_infinite (fun a => N a \/ a = el).
Proof.
  intros; apply countably_infinite_set2_impl_set.
  apply countably_infinite_set_impl_set2 in INF; rdes INF.
  exists (fun a : {a | N a \/ a = el} => let (a,PA) := a in 
            match or2sum PA with | left PA => S (f (exist _ a PA)) | right _ => 0 end).
  exists (fun n => exist (fun a => N a \/ a = el) 
            (proj1_sig (g n)) (or_introl _ (proj2_sig (g n)))).
  split; repeat intro.
  - destruct x as (x & PX), y as (y & PY); simpls.
    destruct (or2sum PX), (or2sum PY); inv H; eauto using proj1_sig_inj.
    by apply INF in H; inv H; eauto using proj1_sig_inj.
  - by inv H; eapply proj1_sig_inj, INF0 in H1.
Qed.

Lemma countably_infinite_minus_elmt: 
  forall A (N: A -> Prop) (INF: countably_infinite N) (el: A),
    countably_infinite (fun a => N a /\ a <> el).
Proof.
  intros; destruct (LEM (N el)) as [NEL | NEL].
  - repeat red in INF; des. 
    set (nel := f (exist _ _ NEL)).
    assert (NOTEL: forall n, proj1_sig (g (S nel + n)) <> el).
      red; intros; cut (S nel + n = nel); [by omega|].
      by rewrite <- INF0 at 1; unfold nel at 2; f_equal; eauto using proj1_sig_inj.

    apply countably_infinite_set2_impl_set; repeat red in INF; des. 
    exists (fun a : sig (fun a => N a /\ a <> el) => 
              f (exist _ (proj1_sig a) (proj1 (proj2_sig a)))).
    exists (fun n => exist (fun a => N a /\ a <> el) _ 
              (conj (proj2_sig (g (S nel+n))) (NOTEL n))).
    split; red; intros.
    + by eapply f_equal in H; rewrite !INF in H; inversion H; eauto using proj1_sig_inj.
    + by inversion H; eapply proj1_sig_inj, f_equal in H1; rewrite !INF0 in H1; omega.

  - replace (fun a => N a /\ a <> el) with N; eauto.
    by extensionality a; apply prop_ext; intuition; subst.
Qed.

Lemma countably_infinite_split_finite: 
  forall A (N: A -> Prop) (INF: countably_infinite N) (L: list A),
    exists N1 N2, 
      <<Nci1: countably_infinite N1>> /\ <<Nci2: countably_infinite N2>>
      /\ <<Nunion: N = nunion N1 N2>> /\ <<Ndisj: ndisj N1 N2>>
      /\ <<Ldisj: forall a (INL: List.In a L) (INN: N2 a), False>>.
Proof.
  induction L; intros.
    by assert (X := countably_infinite_split INF); des; subst; eauto 10.
  des; subst; unnw.
  exists (fun a' => N1 a' \/ (N2 a' /\ a' = a)); exists (fun a' => N2 a' /\ a' <> a).
  split.
  - destruct (LEM (N2 a)).
    + replace (fun a' => N1 a' \/ N2 a' /\ a' = a) with (fun a' => N1 a' \/ a' = a).
        by eauto using countably_infinite_plus_elmt.
      by extensionality a'; apply prop_ext; split; intros; des; subst; eauto.
    + replace (fun a' => N1 a' \/ N2 a' /\ a' = a) with N1; eauto.
      by extensionality a'; apply prop_ext; split; intros; des; subst; eauto.
  - split; eauto using countably_infinite_minus_elmt.
    split.
      by extensionality x; apply prop_ext; unfold nunion; split; intros;
         destruct (LEM (x = a)); des; eauto.
    by split; repeat intro; simpls; des; eauto.
Qed.

(** A product of countably infinite sets is countably infinite. *)

Opaque nat_pair.

Lemma cntinf_prod: forall A B (INFA: countably_infinite_set A) (INFB: countably_infinite_set B),
  countably_infinite_set (A * B).
Proof.
intros; apply countably_infinite_set2_impl_set.
red in INFA, INFB |- *; intros; des.
exists (fun ab => nat_pair (f0 (fst ab), f (snd ab))).
exists (fun n => (g0 n, g 0)).
split; red; intros.
  destruct x, y; eapply nat_pair_inj in H; inv H.
  by rewrite <-(INFA a), <-(INFA a0), H1, <-(INFB b), <-(INFB b0), H2. 
by inv H; rewrite <-(INFA0 x), <-(INFA0 y), H1.
Qed.

(**
   location is countably infinite
*)

Lemma location_cntinf: countably_infinite_set location.
Proof.
  by do 2 exists (@id location); eauto.
Qed.

(**
   typ is countably infinite
*)

Lemma typ_cntinf: countably_infinite_set typ.
Proof.
  set (tyenc := fix tyenc ty := 
    match ty with
      | Tname n => nat_pair (0,n)
      | Tbvar a => nat_pair (1,a)
      | Tunit => nat_pair (2,0)
      | Tint => nat_pair (3,0)
      | Tbool => nat_pair (4,0)
      | Tprod ty1 ty2 => nat_pair (5, nat_pair (tyenc ty1, tyenc ty2))
      | Tsum ty1 ty2 => nat_pair (6, nat_pair (tyenc ty1, tyenc ty2))
      | Tarrow ty1 ty2 => nat_pair (7, nat_pair (tyenc ty1, tyenc ty2))
      | Tref ty => nat_pair (8, tyenc ty)
      | Tmu ty => nat_pair (9, tyenc ty)
      | Tforall ty => nat_pair (10, tyenc ty)
      | Texists ty => nat_pair (11, tyenc ty)
    end).
  set (natembed n := Tname n).

  eapply countably_infinite_set2_impl_set.
  exists tyenc; exists natembed; split; [|by red; intros EQ; intros; inv EQ].
  intro ty; induction ty; intro ty'; destruct ty'; simpl; intros; eapply nat_pair_inj in H; inv H;
    try (by erewrite IHty; eauto);
    try (by eapply nat_pair_inj in H; inv H; erewrite IHty1; eauto; 
            eapply nat_pair_inj in H1; inv H1; erewrite IHty2; eauto).
Qed.

(**
   val is countably infinite
*)

Lemma val_cntinf: countably_infinite_set val.
Proof.
  set (expenc := fix expenc e :=
    match e with
      | Evar a => nat_pair (0, a)
      | Eunit => nat_pair (1, 0)
      | Eint i => nat_pair (2, i)
      | Etrue => nat_pair (3, 0)
      | Efalse => nat_pair (4, 0)
      | Epair e1 e2 => nat_pair (5, nat_pair (expenc e1, expenc e2))
      | Einl e => nat_pair (6, expenc e)
      | Einr e => nat_pair (7, expenc e)
      | Efst e => nat_pair (8, expenc e)
      | Esnd e => nat_pair (9, expenc e)
      | Eloc l => nat_pair (10, l)
      | Eif e1 e2 e3 => nat_pair (11, nat_pair (expenc e1, nat_pair (expenc e2, expenc e3)))
      | Ecase a e1 e2 e3 => nat_pair (12, nat_pair (a, nat_pair (expenc e1, nat_pair (expenc e2, expenc e3))))
      | Eroll e => nat_pair (13, expenc e)
      | Eunroll e => nat_pair (14, expenc e)
      | Efix f a e => nat_pair (15, nat_pair (f, nat_pair (a, expenc e)))
      | Eapp e1 e2 => nat_pair (16, nat_pair (expenc e1, expenc e2))
      | Egen e => nat_pair (17, expenc e)
      | Einst e => nat_pair (18, expenc e)
      | Epack e => nat_pair (19, expenc e)
      | Eunpack a e1 e2 => nat_pair (20, nat_pair (a, nat_pair (expenc e1, expenc e2)))
      | Enew e => nat_pair (21, expenc e)
      | Eread e => nat_pair (22, expenc e)
      | Ewrite e1 e2 => nat_pair (23, nat_pair (expenc e1, expenc e2))
      | Erefeq e1 e2 => nat_pair (24, nat_pair (expenc e1, expenc e2))
    end).
  set (valenc := fix valenc v := 
    match v with
      | Vvar a  => nat_pair (0,a)
      | Vunit => nat_pair (1,0)
      | Vint i => nat_pair (2,i)
      | Vtrue => nat_pair (3,0)
      | Vfalse => nat_pair (4,0)
      | Vpair v1 v2 => nat_pair (5, nat_pair (valenc v1, valenc v2))
      | Vinl v => nat_pair (6, valenc v)
      | Vinr v => nat_pair (7, valenc v)
      | Vroll v => nat_pair (8, valenc v)
      | Vfix f a e => nat_pair (9, nat_pair (f, nat_pair (a, expenc e)))
      | Vgen e => nat_pair (10, expenc e)
      | Vpack v => nat_pair (11, valenc v)
      | Vloc l => nat_pair (12, l)
    end).
  set (natembed n := Vvar n).

  assert (EINJ: injective _ _ expenc).
    intro e; induction e; intro e'; destruct e'; simpl; intros; eapply nat_pair_inj in H; inv H;
      try (by try (eapply nat_pair_inj in H; inv H); try (eapply nat_pair_inj in H1; inv H1); 
              try (eapply nat_pair_inj in H3; inv H3); try (eapply nat_pair_inj in H5; inv H5);
              try (erewrite IHe; eauto); try (erewrite IHe1; eauto); 
              try (erewrite IHe2; eauto); try (erewrite IHe3; eauto)).
  
  eapply countably_infinite_set2_impl_set; red in EINJ.
  exists valenc; exists natembed; split; [|by red; intros EQ; intros; inv EQ].
  intro v; induction v; intro v'; destruct v'; simpl; intros; eapply nat_pair_inj in H; inv H;
    try (by try (eapply nat_pair_inj in H; inv H); try (eapply nat_pair_inj in H1; inv H1); 
            try (eapply nat_pair_inj in H3; inv H3); 
            try (erewrite IHv; eauto); try (erewrite IHv1; eauto); try (erewrite IHv2; eauto);
            try (erewrite (EINJ _ _ H); eauto); try (erewrite (EINJ _ _ H3); eauto)). 
Qed.

