Require Import Program Bool List Arith basis lang.
Require Import Omega FunctionalExtensionality Compare_dec Le Image List Mult.
Require Import Classical IndefiniteDescription ChoiceFacts.
Require Import vgtac.

Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** ** Classical Logic *)
(*----------------------------------------------------------------------------*)

(** We assume classical logic, functional and propositional extensionality. *)

Lemma prop_ext : ClassicalFacts.prop_extensionality.
Proof.
  red; intros; exploit (@Extensionality_Ensembles () (fun _ => A) (fun _ => B)).
  - by split; red; intros; unfold In in *; intuition.
  - intros EQ; change A with ((fun _ => A) ()); change B with ((fun _ => B) ()).
    by rewrite EQ.
Qed.

Lemma proof_irr: ClassicalFacts.proof_irrelevance.
Proof. exact (ClassicalFacts.ext_prop_dep_proof_irrel_cic prop_ext). Qed.

Lemma proj1_sig_inj: forall A (N : A -> Prop) (a a': sig N) (EQ: `a = `a'), a = a'.
Proof.
intros; destruct a, a'; simpls; subst; f_equal; eauto using proof_irr.
Qed.

Lemma sig_eta: forall A (P: A -> Prop) (x : sig P) pf, exist P (`x) pf = x.
Proof. by intros; destruct x; simpls; apply proj1_sig_inj. Qed.

Lemma LEM: forall (P: Prop), {P} + {~P}.
Proof.
intros; eapply constructive_definite_descr_excluded_middle; eauto using classic.
repeat intro; eapply constructive_indefinite_description in H.
by destruct H as (a & H); destruct H; eauto.
Qed.

Lemma or2sum: forall P Q (OR: P \/ Q), {P} + {Q}.
Proof.
intros; destruct (LEM P); [by left|]; destruct (LEM Q); [by right|].
by exfalso; intuition.
Qed.

Ltac gen_cid :=
  match goal with [|- context[@constructive_indefinite_description ?A ?B ?C]] =>
    generalize (@constructive_indefinite_description A B C)
  end.

Ltac gen_eqrefl :=
  match goal with [|-context[@eq_refl ?A ?B]] => 
    generalize (@eq_refl A B) 
  end.

(*----------------------------------------------------------------------------*)
(** ** Miscellaneous 'library' definitions and lemmas *)
(*----------------------------------------------------------------------------*)

Hint Unfold In.

Definition swap {A B} (x : A * B) := (snd x, fst x).

Lemma list_minus_emp_spec :
  forall A myeq l1 l2,
    (list_minus (A:=A) myeq l1 l2 = nil) <-> (forall x, In x l1 -> In x l2).
Proof.
  intros; induction l1 as [|a l1 IH]; simpl; intros; try tauto.
  generalize (list_mem_spec myeq a l2); destruct list_mem;
  intuition; congruence. 
Qed.

Lemma list_minus_self: forall A eqx l l', @list_minus A eqx l (l'++l) = nil.
Proof. 
  induction l; simpl; intros; eauto.
  assert (X:= list_mem_spec eqx a (l'++a::l)).
  destruct (list_mem _ _ _); [|by exfalso; apply X; apply in_or_app; simpl; eauto].
  setoid_rewrite (@app_assoc _ _ (a::nil)); eauto.
Qed.

Lemma list_minus2_red : forall A B eqx l l1 l2,
  list_minus2 eqx (@list_minus2 A B eqx l l2) (list_minus eqx l1 l2) = 
  list_minus2 eqx (list_minus2 eqx l l2) l1.
Proof.
  induction l as [|a l IH]; simpl; intros; auto.
  destruct a; destruct (list_mem _ _ _) as [] _eqn: X; simpls; auto.
  rewrite list_mem_minus, X, IH; destruct (list_mem _ _ l1); trivial.
Qed.

Lemma list_minus2_len: forall A B eqx l r, length (@list_minus2 A B eqx l r) <= length l.
Proof.
  induction l; s; i; eauto.
  destruct (list_mem _ _ _); s; eauto.
  specialize (IHl r); omega.
Qed.

Lemma list_minus_nil_r: forall A eqx l, @list_minus A eqx l nil = l.
Proof. intros; induction l; simpl; try rewrite IHl; auto. Qed.

Lemma list_minus2_nil_r : forall A B eqx l, @list_minus2 A B eqx l nil = l.
Proof. intros; induction l; simpl; try rewrite IHl; auto. Qed.

Lemma app_nil_invl: forall A (l1 l2: list A), l1 = [] -> l2 = [] -> l1 ++ l2 = [].
Proof. intros; subst; eauto. Qed.

Lemma app_nil_invr: forall A (l1 l2: list A), l1 ++ l2 = [] -> l1 = [] /\ l2 = [].
Proof. intros; destruct l1; destruct l2; eauto; inv H. Qed.

Lemma list_assoc_map: forall A B C myeq (f: B -> C) l a b
    (IN: @list_assoc A B myeq a l = Some b),
  list_assoc myeq a (map (fun x => (fst x, f (snd x))) l) = Some (f b).
Proof.
  induction l; ss.
  i; destruct a; destruct (myeq _ _); subst; inv IN; eauto.
Qed.

Lemma list_assoc_map_inv: forall A B C myeq (f: B -> C) l a c
    (IN: @list_assoc A C myeq a (map (fun x => (fst x, f (snd x))) l) = Some c),
  exists b, f b = c /\ list_assoc myeq a l = Some b.
Proof.
  induction l; ss.
  i; destruct a; destruct (myeq _ _); subst; inv IN; eauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * Type substitutions for type variables *)
(*----------------------------------------------------------------------------*)

Lemma lc_mon :
  forall ty n m (LT: n <= m), lc_typ n ty -> lc_typ m ty.
Proof.
  induction ty; simpl in *; intros; intuition eauto using le_n_S; omega.
Qed.

Lemma lc_tsubst1 :
  forall tyA ty x, 
    lc_typ 0 tyA -> 
    lc_typ (S x) ty -> 
    lc_typ x (tsubst1 x tyA ty).
Proof.
  induction ty; simpl in *; intros; intuition eauto. 
  destruct (nat_compare_spec bb x); simpl; eauto using lc_mon, le_0_n.
  omega. 
Qed.

Lemma lc_tsubst1D:
  forall tyA ty n,
    lc_typ n (tsubst1 n tyA ty) ->
    lc_typ (S n) ty.
Proof.
  induction ty; simpl; intros; f_equal; intuition auto.
  destruct (nat_compare_spec bb n); simpl in *; omega.
Qed.

Lemma tsubst1_id : 
  forall tyA ty x, lc_typ x ty -> tsubst1 x tyA ty = ty.
Proof.
  induction ty; intross; des; f_equal; eauto.
  destruct (nat_compare_spec bb x); try (cut False; [|omega]); try done.
Qed.

Lemma in_ftn_tsubst1_mon : 
  forall aa tyA ty n (IN: In aa (ftn_typ ty)), In aa (ftn_typ (tsubst1 n tyA ty)).
Proof.
  induction ty; intross; try eapply in_app_or in IN; des;
  try specialize (IHty1 _ IN); try specialize (IHty2 _ IN); des;
  eauto using in_or_app. 
Qed.

Lemma ftn_tsubst1_mon_closed : 
  forall tyA ty n (CLO: ftn_typ (tsubst1 n tyA ty) = nil), ftn_typ ty = nil.
Proof.
  induction ty; intross; try eapply app_eq_nil in CLO; des;
  try specialize (IHty1 _ CLO); try specialize (IHty2 _ CLO0);
  try rewrite IHty1, IHty2; eauto.
Qed.

Lemma in_ftn_tsubst1 : 
  forall aa tyA ty n (IN: In aa (ftn_typ (tsubst1 n tyA ty))),
    In aa (ftn_typ tyA) \/ In aa (ftn_typ ty).
Proof.
  induction ty; intross; try eapply in_app_or in IN; des;
  try specialize (IHty1 _ IN); try specialize (IHty2 _ IN); des;
  eauto using in_or_app. 
  destruct nat_compare; vauto.
Qed.

Lemma ftn_tsubst1_closed : 
  forall tyA ty n (CLO: ftn_typ tyA = nil) (CLO': ftn_typ ty = nil),
    ftn_typ (tsubst1 n tyA ty) = nil.
Proof.
  induction ty; intross; try eapply app_eq_nil in CLO'; des; 
  try specialize (IHty1 n CLO CLO'); try specialize (IHty2 n CLO CLO'0);
  try rewrite IHty1, IHty2; eauto.
  destruct nat_compare; vauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * Type substitutions for free type names *)
(*----------------------------------------------------------------------------*)

Definition tsubst_lclosed (tsub: list (typename * typ)) := 
  forall x (IN: In x tsub), lc_typ 0 (snd x).

Lemma lc_tsubst:
  (* forall d (CLO: tsubst_lclosed d)  *)
  forall d (CLO: forall x (IN: In x d), lc_typ 0 (snd x))
    ty n (H: lc_typ n ty),
    lc_typ n (tsubst_typ d ty).
Proof.
  induction ty; intross; des; auto.
  pose proof (list_assoc_spec eq_typename aa d).
  destruct list_assoc; eauto.
  apply CLO in H0; simpl in *; eauto using lc_mon, le_0_n.
Qed.

Lemma tsubst_lc: forall s ty n (LC: lc_typ n (tsubst_typ s ty)), lc_typ n ty.
Proof. by intros s ty; induction ty; intros; simpls; des; eauto. Qed.

Lemma tsubst_typ_tsubst1 : 
  forall d tyA ty x
    (CLO: tsubst_lclosed d),
    lc_typ (S x) ty ->
    tsubst_typ d (tsubst1 x tyA ty) = tsubst1 x (tsubst_typ d tyA) (tsubst_typ d ty).
Proof.
  induction ty; intross; des; f_equal; eauto.
  - pose proof (list_assoc_spec eq_typename aa d);
    destruct list_assoc; simpls.
    by apply CLO in H0; simpl in *; rewrite tsubst1_id; eauto using lc_mon, le_0_n.
  - by destruct nat_compare; simpls.
Qed.

Lemma tsubst_typ_disj : 
  forall x sub (H: list_disj eq_typename (map (@fst _ _) sub) (ftn_typ x)), tsubst_typ sub x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve tsubst_typ_disj : subst_lemmas.

Lemma tsubst_typ_closed : 
  forall x (H: ftn_typ x = nil) sub, tsubst_typ sub x = x.
Proof. intros; apply tsubst_typ_disj; rewrite H; constructor. Qed.
Hint Resolve tsubst_typ_closed : subst_lemmas.
Hint Resolve (fun x H sub => eq_sym (tsubst_typ_closed x H sub)) : subst_lemmas. 

Lemma tsubst_typ_nil : forall x, tsubst_typ nil x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve tsubst_typ_nil : subst_lemmas.
Hint Rewrite tsubst_typ_nil : subst_rw_lemmas.

Lemma tsubst_typ_cons : 
  forall x y z l (CLOSED: ftn_typ y = nil),
    tsubst_typ ((x,y)::l) z = tsubst_typ l (tsubst_typ (cons (x,y) nil) z). 
Proof. induction z; subst_tac'. Qed.
Hint Resolve tsubst_typ_cons : subst_lemmas.

Lemma tsubst_typ_remdup :
  forall x l, tsubst_typ (list_remdup2 eq_typename l) x = tsubst_typ l x.
Proof. induction x; simpl; intros; autorewrite with subst_rw_lemmas; f_equal; auto. Qed.
Hint Rewrite tsubst_typ_remdup : subst_rw_lemmas.

Lemma tsubst_typ_cons_minus2: 
  forall x e' l e,
    tsubst_typ ((x, e') :: list_minus2 eq_typename l (x :: nil)) e
    = tsubst_typ ((x, e') :: l) e.
Proof.
  intros.
  rewrite <- (tsubst_typ_remdup _ (_ :: list_minus2 _ _ _)).
  rewrite <- (tsubst_typ_remdup _ (_ :: l)); simpl.
  rewrite <- list_minus2_remdup2, <- list_minus2_app_r.
  rewrite list_minus2_eq with (l2 := x :: nil); simpl; tauto.
Qed.

Lemma tsubst_typ_eq:
  forall l l' e,
    list_remdup2 eq_typename l = list_remdup2 eq_typename l' ->
    tsubst_typ l e = tsubst_typ l' e.
Proof. intros; rewrite <- tsubst_typ_remdup, H, tsubst_typ_remdup; auto. Qed.

Lemma tsubst_typ_nin:
   forall aa tyA d ty, 
     ~ In aa (ftn_typ ty) ->
   tsubst_typ ((aa, tyA) :: d) ty = tsubst_typ d ty.
Proof.
    induction ty; intross; f_equal; eauto;
    repeat match goal with
      | H : ~ In _ _ |- ~ In _ _ => intro; destruct H
      | |- In _ (_ ++ _) => apply in_or_app; auto
      | |- In _ (list_minus _ _ _) => apply in_list_minusI; simpl; auto
      | H: _ |- _ => apply H 
      | |- context[if (eq_typename ?x ?y) then _ else _] => destruct (eq_typename x y); subst; try done
    end; tauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * Exp substitutions *)
(*----------------------------------------------------------------------------*)

Notation "sub '\s\' rm" := (list_minus2 eq_termvar sub rm) 
                             (at level 50, no associativity).
Notation "fvs '\f\' rm" := (list_minus eq_termvar fvs rm) 
                             (at level 50, no associativity).

Definition subst := list (termvar * exp).
Definition substv := list (termvar * val).
Definition sv2e (sub: substv) : subst := 
  map (fun x => (fst x, exp_of_val (snd x))) sub.

Fixpoint subst_closed (sub : list (termvar * exp)) :=
  match sub with
    | nil => True
    | h :: l => fv_exp (snd h) = nil /\ subst_closed l
  end.

Definition subst_closed_i (sub : subst) := 
  forall x e (IN: In (x,e) sub), fv_exp e = nil.
Hint Unfold subst_closed_i.
Arguments subst_closed_i sub /.

Lemma sv2e_cons: forall x ve sub, (x,e_of ve):: sv2e sub = sv2e ((x,v_of ve)::sub).
Proof. intros; destruct ve; subst; eauto. Qed.

Lemma subst_closed_eq: subst_closed = subst_closed_i.
Proof.
  extensionality sub; apply prop_ext; unfold subst_closed_i; induction sub; [s; tauto|].
  destruct a; s; split; i.
  - des; eauto; inv IN.
  - des; split; eauto.
Qed.

Lemma subst_closed_app : 
  forall l l',
    subst_closed (l ++ l') <-> subst_closed l /\ subst_closed l'.
Proof. intros; induction l; simpl; tauto. Qed.

Lemma subst_closed_minus2 : 
  forall l l',
    subst_closed l ->
    subst_closed (l \s\ l').
Proof. induction l; simpl; intros; [|destruct list_mem]; simpl; intuition. Qed.
Hint Resolve subst_closed_minus2 : subst_lemmas.

Lemma subst_exp_disj : 
  forall (x: exp) sub (H: list_disj eq_termvar (map (@fst _ _) sub) (fv_exp x)), subst_exp sub x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve subst_exp_disj : subst_lemmas.

Lemma subst_exp_closed : 
  forall (x: exp) (H: fv_exp x = nil) sub, subst_exp sub x = x.
Proof. intros; apply subst_exp_disj; rewrite H; constructor. Qed.
Hint Resolve subst_exp_closed : subst_lemmas.
Hint Resolve (fun x H sub => eq_sym (subst_exp_closed x H sub)) : subst_lemmas. 

Lemma subst_exp_nil : forall (x: exp), subst_exp nil x = x.
Proof. induction x; subst_tac. Qed. 
Hint Resolve subst_exp_nil : subst_lemmas.
Hint Rewrite subst_exp_nil : subst_rw_lemmas.

Lemma subst_exp_cons : 
  forall x (e' e: exp) l (CLOSED: fv_exp e' = nil),
    subst_exp ((x,e')::l) e = subst_exp l (subst_exp (cons (x,e') nil) e). 
Proof. induction e; subst_tac'. Qed.
Hint Resolve subst_exp_cons : subst_lemmas.

Lemma subst_exp_remdup :
  forall x l, @subst_exp (list_remdup2 eq_termvar l) x = subst_exp l x.
Proof. induction x; simpl; intros; autorewrite with subst_rw_lemmas; f_equal; auto. Qed.
Hint Rewrite subst_exp_remdup : subst_rw_lemmas.

Lemma subst_exp_app :
  forall l (C: subst_closed l) l' e,
  subst_exp (l ++ l') e = subst_exp l' (subst_exp l e). 
Proof.
  induction l as [|[? ?] ? IH]; intros; simpl in *;
   [ autorewrite with subst_rw_lemmas
   | rewrite subst_exp_cons, IH, <- subst_exp_cons]; tauto.
Qed.

Lemma subst_exp_app_flip :
  forall l l' (D: list_disj eq_termvar (map (@fst _ _) l) (map (@fst _ _) l')) (e: exp),
  subst_exp (l ++ l') e = subst_exp (l' ++ l) e. 
Proof.
  intros; revert l l' D; induction e; intros; simpl; 
    rewrite ?list_minus2_app, ?(list_assoc_flip _ _ _ D); subst_tac'.
Qed.

Lemma subst_exp_app_disj :
  forall l' (C: subst_closed l') l
   (D: list_disj eq_termvar (map (@fst _ _) l) (map (@fst _ _) l')) e,
  subst_exp l (subst_exp l' e) = subst_exp (l ++ l') e.
Proof.
  intros; rewrite <- subst_exp_app, <- subst_exp_app_flip; auto.
Qed.

Lemma subst_exp_comm :
  forall l (C: subst_closed l) l' (C': subst_closed l')
   (D: list_disj eq_termvar (map (@fst _ _) l) (map (@fst _ _) l')) e,
    subst_exp l (subst_exp l' e) = (subst_exp l' (subst_exp l e)).
Proof.
  intros; rewrite <- !subst_exp_app, (subst_exp_app_flip _ _ D); auto.
Qed.

Lemma fv_subst_exp :
  forall l (C: subst_closed l) e,
  fv_exp (subst_exp l e) = fv_exp e \f\ map (@fst _ _) l.
Proof.
  intros; revert l C; induction e; simpl; intros; trivial; rewrite ?list_minus_app;
  repeat match goal with 
           | |- _ => solve [auto]
           | |- _ ++ _ = _ ++ _ => f_equal
           | H : _ |- _ => solve[rewrite H, list_map_fst_minus2, list_minus_comm, 
                                         list_minus_red; auto with subst_lemmas] end.
  induction l as [|[]]; simpl in *; [|destruct (eq_termvar)]; intuition. 
Qed.

Lemma subst_exp_cons_minus2: 
  forall x e' l (e: exp),
    subst_exp ((x, e') :: (l \s\ (x :: nil))) e
    = subst_exp ((x, e') :: l) e.
Proof.
  intros.
  rewrite <- (subst_exp_remdup _ (_ :: list_minus2 _ _ _)).
  rewrite <- (subst_exp_remdup _ (_ :: l)); simpl.
  rewrite <- list_minus2_remdup2, <- list_minus2_app_r.
  rewrite list_minus2_eq with (l2 := x :: nil); simpl; tauto.
Qed.

Lemma subst_exp_eq:
  forall l l' (e: exp),
    list_remdup2 eq_termvar l = list_remdup2 eq_termvar l' ->
    subst_exp l e = subst_exp l' e.
Proof. intros; rewrite <- subst_exp_remdup, H, subst_exp_remdup; auto. Qed.

Lemma subst_exp_app_swap: forall (e: exp) sub sub',
  subst_exp ((sub' \s\ map fst sub) ++ sub) e =
  subst_exp (sub ++ sub') e.
Proof.
  intros; rewrite subst_exp_app_flip.
  - setoid_rewrite <-subst_exp_remdup; f_equal.
    revert sub'; induction sub; simpl; intros.
    + by rewrite list_minus2_nil_r.
    + destruct a; simpl.
      rewrite <-(IHsub sub'), !list_minus2_remdup2, !list_minus2_app.
      setoid_rewrite list_minus2_comm; setoid_rewrite <-list_minus2_red.
      by setoid_rewrite (list_minus_app _ [t]); setoid_rewrite (list_minus_self _ _ []).
  - rewrite <-list_minus_map_fst; apply list_disj_minus_l.
    by setoid_rewrite (list_minus_self _ _ []).
Qed.

Lemma fv_cont_appl: forall k1 k2, @fv_cont k1 = [] -> fv_cont k2 = [] -> 
  fv_cont (k1 ++ k2) = [].
Proof. i; unfold fv_cont in *; rewrite fold_right_app; rewrite H0, H; eauto. Qed.

Lemma fv_cont_appr: forall k1 k2, @fv_cont (k1 ++ k2) = [] ->
  fv_cont k1 = [] /\ fv_cont k2 = [].
Proof. 
  unfold fv_cont in *; induction k1; eauto; intros.
  ss; apply app_nil_invr in H; des.
  by apply IHk1 in H0; des; rewrite H, H0, H1.
Qed.

