Require Import Program Bool List Arith basis lang.
Require Import Omega FunctionalExtensionality Compare_dec Le Image List Mult.
Require Import Classical IndefiniteDescription ChoiceFacts.
Require Import vgtac.

Set Implicit Arguments.
Unset Strict Implicit.

(*----------------------------------------------------------------------------*)
(** ** Classical Logic *)
(*----------------------------------------------------------------------------*)

(** We assume classical logic, functional and propositional extensionality. *)

Definition cid := constructive_indefinite_description.

Lemma prop_ext : ClassicalFacts.prop_extensionality.
Proof.
  red; intros; exploit (@Extensionality_Ensembles () (fun _ => A) (fun _ => B)).
  - by split; red; intros; unfold In in *; intuition.
  - intros EQ; change A with ((fun _ => A) ()); change B with ((fun _ => B) ()).
    by rewrite EQ.
Qed.

Lemma proof_irr: ClassicalFacts.proof_irrelevance.
Proof. exact (ClassicalFacts.ext_prop_dep_proof_irrel_cic prop_ext). Qed.

Lemma proj1_sig_inj: forall A (N : A -> Prop) (a a': sig N) (EQ: `a = `a'), a = a'.
Proof.
intros; destruct a, a'; simpls; subst; f_equal; eauto using proof_irr.
Qed.

Lemma sig_eta: forall A (P: A -> Prop) (x : sig P) pf, exist P (`x) pf = x.
Proof. by intros; destruct x; simpls; apply proj1_sig_inj. Qed.

Lemma sig_eq : 
  forall A (P: A -> Prop) (x x' : sig P), proj1_sig x = proj1_sig x' -> x = x'.
Proof. destruct x, x'; simpl; intro; subst; f_equal; apply proof_irr. Qed.

Lemma LEM: forall (P: Prop), {P} + {~P}.
Proof.
intros; eapply constructive_definite_descr_excluded_middle; eauto using classic.
repeat intro; eapply cid in H.
by destruct H as (a & H); destruct H; eauto.
Qed.

Lemma or2sum: forall P Q (OR: P \/ Q), {P} + {Q}.
Proof.
intros; destruct (LEM P); [by left|]; destruct (LEM Q); [by right|].
by exfalso; intuition.
Qed.

Ltac gen_cid :=
  match goal with [|- context[@cid ?A ?B ?C]] =>
    generalize (@cid A B C)
  end.

Ltac gen_eqrefl :=
  match goal with [|-context[@eq_refl ?A ?B]] => 
    generalize (@eq_refl A B) 
  end.

(*----------------------------------------------------------------------------*)
(** * eq_termvar equalities *)
(*----------------------------------------------------------------------------*)

Lemma eqtmv_true: forall x, eq_termvar x x = left _ refl_equal.
Proof.
  by intros; destruct eq_termvar; [f_equal; apply UIP|].
Qed.

Lemma eqtmv_false: forall x y (NEQ: ~ x = y), eq_termvar x y = right _ NEQ.
Proof.
  i; destruct eq_termvar; [exfalso; eauto | f_equal; apply proof_irrelevance].
Qed.

(*----------------------------------------------------------------------------*)
(** * list_flat *)
(*----------------------------------------------------------------------------*)

Fixpoint list_flat A (ll: list (list A)) : list A :=
  match ll with
  | nil => nil
  | l :: ll' => l ++ list_flat ll'
  end.

Lemma list_flat_in: forall A l (a: A) ll
  (INl:In l ll) (INa: In a l), In a (list_flat ll).
Proof.
  induction ll; [done|].
  s; i; apply in_or_app; des; subst; eauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * ftn_typL *)
(*----------------------------------------------------------------------------*)

Lemma ftn_typ_finite: 
  forall ty, forall a (IN: ftn_typ ty a), In a (ftn_typL ty).
Proof.
  induction ty; s; i; try apply in_or_app; des; eauto.
Qed.

Lemma ftn_typL_in: forall ty a (IN: ftn_typ ty a), In a (ftn_typL ty).
Proof. 
  induction ty; ss; ii; des; eauto using in_or_app.
Qed.

Lemma ftn_typL_in_iff: forall ty a, ftn_typ ty a <-> In a (ftn_typL ty).
Proof. 
  induction ty; ss; ii; des; rewrite ?in_app_iff; firstorder. 
Qed.

(*----------------------------------------------------------------------------*)
(** * Type substitutions *)
(*----------------------------------------------------------------------------*)

Definition tsubst_lclosed (tsub: typename -> typ) := 
  forall x, lc_typ 0 (tsub x).
Hint Unfold tsubst_lclosed.

Lemma lc_mon :
  forall ty n m (LT: n <= m), lc_typ n ty -> lc_typ m ty.
Proof.
  induction ty; simpl in *; intros; intuition eauto using le_n_S; omega.
Qed.

Lemma lc_tsubst:
  forall d (CLO: tsubst_lclosed d)
    ty n (H: lc_typ n ty),
    lc_typ n (tsubst_name d ty).
Proof.
  induction ty; intross; des; eauto using lc_mon, le_0_n.
Qed.

Lemma tsubst_lc: forall s ty n (LC: lc_typ n (tsubst_name s ty)), lc_typ n ty.
Proof. by intros s ty; induction ty; intros; simpls; des; eauto. Qed.

Lemma tsubst_bvar_id : 
  forall tyA ty x, lc_typ x ty -> tsubst_bvar x tyA ty = ty.
Proof.
  induction ty; intross; des; f_equal; eauto.
  destruct (nat_compare_spec bb x); try (cut False; [|omega]); try done.
Qed.

Lemma tsubst_name_bvar:
  forall t ty b dd (LC: forall x, lc_typ 0 (dd x)) , 
    tsubst_name dd (tsubst_bvar b t ty) =
    tsubst_bvar b (tsubst_name dd t) (tsubst_name dd ty).
Proof.
  induction ty; i; s; f_equal; eauto.
  2: by destruct nat_compare.
  rewrite tsubst_bvar_id; eauto using lc_mon with arith.
Qed.

Lemma tsubnam_agree : 
  forall (ty: typ) d d' (AGREE: forall x (IN: ftn_typ ty x), d x = d' x),
  tsubst_name d ty = tsubst_name d' ty.
Proof.
  induction ty; s; i; 
  try erewrite IHty; try erewrite IHty1; try erewrite IHty2; eauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * Exp substitutions *)
(*----------------------------------------------------------------------------*)

Definition subst_closed (sub : termvar -> exp) :=
  forall x a, ~ fv_exp (sub x) a.
Hint Unfold subst_closed.

Definition subst_weakly_closed (sub : termvar -> exp) :=
  forall x a (FV: fv_exp (sub x) a), a = x.
Hint Unfold subst_weakly_closed.

Lemma sext_sid: forall x,
  sext x (Evar x) sid = sid.
Proof.
  i; extensionality a; unfold sext; destruct (eq_termvar _ _); eauto.
Qed.

Lemma sext_shadow: forall A sub x (a a': A),
  sext x a' (sext x a sub) = sext x a' sub.
Proof.
  i; extensionality y; unfold sext; destruct (eq_termvar _ _); eauto.
Qed.

Lemma sext_swap_var: forall x x' (NEQ: x <> x') sub ex,
  sext x' (Evar x') (sext x ex sub) = sext x ex (sext x' (Evar x') sub).
Proof.
  i; extensionality y; unfold sext; destruct (eq_termvar y x'); eauto.
  by subst; destruct (eq_termvar _ _); subst; eauto.
Qed.

Lemma sext_map: forall A B (f: A -> B) sub x a,
  f <*> (sext x a sub) = sext x (f a) (f <*> sub).
Proof.
  i; extensionality y; unfold compose, sext; destruct eq_termvar; subst; eauto.
Qed.

Lemma subclo_weakly: forall sub,
  subst_closed sub -> subst_weakly_closed sub.
Proof. by ii; apply H in FV. Qed.

Lemma weakclo_sext: forall sub x ex
    (CLOSED : subst_weakly_closed sub)
    (CLOEX: forall a, fv_exp ex a -> a = x),
  subst_weakly_closed (sext x ex sub).
Proof.
  unfold sext; ii; destruct (eq_termvar _ _); subst; eauto.
Qed.

Lemma subexp_agree : 
  forall (e: exp) sub sub' (H: forall x (IN: fv_exp e x), sub x = sub' x),
  subst_exp sub e = subst_exp sub' e.
Proof.
  induction e; s; i
  ; try erewrite IHe; try erewrite IHe1; try erewrite IHe2; try erewrite IHe3
  ; eauto 10.
  - i; unfold sext; destruct (eq_termvar _ _); eauto 10.
  - i; unfold sext; destruct (eq_termvar _ _); eauto 10.
  - i; unfold sext; destruct (eq_termvar x0 x); eauto 10.
    destruct (eq_termvar _ _); eauto 10.
  - i; unfold sext; destruct (eq_termvar _ _); eauto 10.
  - i; unfold sext; destruct (eq_termvar _ _); eauto 10.
Qed.

Lemma subexp_sid : forall (e: exp), subst_exp sid e = e.
Proof. 
  induction e; s; i; try rewrite !sext_sid
  ; try erewrite IHe; try erewrite IHe1; try erewrite IHe2; try erewrite IHe3
  ; eauto 10.
Qed.

Lemma subexp_id_closed : 
  forall (e: exp) (H: forall x, ~fv_exp e x) sub, subst_exp sub e = e.
Proof. 
  i; rewrite <-(subexp_sid e) at 2; apply subexp_agree.
  i; exfalso; eauto.
Qed.

Lemma subexp_closed_weakly: forall e sub x
    (WF: wf_exp e)
    (SUB: subst_weakly_closed sub)
    (CLO: forall a, ~ fv_exp (sub x) a),
  ~ fv_exp (subst_exp sub e) x.
Proof.
  unfold subst_weakly_closed; induction e; i; ss; try by des; eauto.
  - assert (FH0:=FH); apply SUB in FH0; subst; eauto.
  - des; [eauto| |].
    + eapply IHe2 in FH0; [eauto|eauto| |]
      ; unfold sext; i; destruct eq_termvar; subst; eauto.
    + eapply IHe3 in FH0; [eauto|eauto| |]
      ; unfold sext; i; destruct eq_termvar; subst; eauto.
  - des; eapply IHe in FH1; [eauto|eauto| |]
    ; unfold sext; i; repeat (destruct eq_termvar; subst; eauto).
  - des; [eauto|].
    des; eapply IHe2 in FH0; [eauto|eauto| |]
    ; unfold sext; i; destruct eq_termvar; subst; eauto.
  - des; eapply IHe in FH0; [eauto|eauto| |]
    ; unfold sext; i; repeat (destruct eq_termvar; subst; eauto).
Qed.

Lemma subexp_closed: forall e sub
    (WF: wf_exp e)
    (CLO: subst_closed sub),
  forall x, ~ fv_exp (subst_exp sub e) x.
Proof.
  i; eapply subexp_closed_weakly in FH; eauto.
  by ii; eapply CLO in FV.
Qed.

Local Ltac sext_case x0 x := 
  let EQ := fresh "EQ" in let NEQ := fresh "NEQ" in let X := fresh "X" in
  (destruct (eq_termvar x0 x) as [EQ|NEQ] eqn: X; subst
   ; [rewrite !sext_shadow, !sext_sid, !subexp_sid
     |rewrite !(sext_swap_var NEQ), !sext_sid]; eauto using weakclo_sext).

Lemma subexp_sext_r: forall e sub x ex (CLOSED: forall a, ~ fv_exp ex a),
  subst_exp (sext x ex sub) e = subst_exp sub (subst_exp (sext x ex sid) e).
Proof.
  induction e; s; i
  ; try by try rewrite IHe; try rewrite IHe1
           ; try rewrite IHe2; try rewrite IHe3; eauto.
  - unfold sext; destruct eq_termvar; [rewrite subexp_id_closed|]; eauto.
  - f_equal; eauto; sext_case x0 x.
  - f_equal; sext_case x0 f; sext_case x0 x.
  - f_equal; eauto; sext_case x0 x.
  - f_equal; sext_case x0 x.
Qed.

Lemma subexp_sext_l: forall e sub x ex (CLOSED: subst_weakly_closed sub),
   subst_exp (sext x ex sub) e = 
   subst_exp (sext x ex sid) (subst_exp (sext x (Evar x) sub) e).
Proof.
  induction e; s; i
  ; try by try rewrite IHe; try rewrite IHe1
           ; try rewrite IHe2; try rewrite IHe3; eauto.
  - unfold sext; destruct (eq_termvar _ _) eqn: X.
    + subst; s; rewrite X; eauto.
    + erewrite subexp_agree; [by rewrite subexp_sid|].
      i; apply CLOSED in IN; subst; rewrite X; eauto.
  - f_equal; eauto; sext_case x0 x.
  - f_equal; sext_case x0 f; sext_case x0 x.
  - f_equal; eauto; sext_case x0 x.
  - f_equal; eauto; sext_case x0 x.
Qed.

Lemma subexp_sext_l2: forall e sub f ef x ex
    (CLOSUB: subst_weakly_closed sub)
    (CLOEX: forall a, ~ fv_exp ex a)
    (CLOEF: forall a, ~ fv_exp ef a),
  subst_exp (sext x ex (sext f ef sub)) e =
  subst_exp (sext x ex (sext f ef sid)) 
    (subst_exp (sext x (Evar x) (sext f (Evar f) sub)) e).
Proof.
  i; destruct (eq_termvar f x) as [EQ|NEQ] eqn: X; subst.
  - rewrite !sext_shadow, <- subexp_sext_l; eauto.
  - Local Ltac _auto := by eapply weakclo_sext; ss; i; exfalso; eauto.
    rewrite subexp_sext_l; [|by _auto].
    rewrite sext_swap_var; [|by eauto].
    rewrite (subexp_sext_l e); [|by _auto].
    rewrite sext_swap_var; [|by eauto].
    rewrite (@subexp_sext_l _ (sext f ef sid)); [|by _auto].
    rewrite (sext_swap_var NEQ sid), sext_sid; eauto.
Qed.

(*----------------------------------------------------------------------------*)
(** * Tactic: Expression To Val *)
(*----------------------------------------------------------------------------*)

Ltac exp2val_i := let TMP := fresh "TMP" in
 rewrite VoE_pf; 
 try (match goal with [|-context[exp_of_val(v_of ?v)]] => 
        set (TMP := exp_of_val(v_of v)); exp2val_i
      end; unfold TMP; clear TMP).
Ltac exp2val := exp2val_i; simpl v_of.
